import datetime
import gzip
import pickle
from typing import Any, Callable, Hashable, Iterable, Type
from uuid import UUID

from gen_epix.fastapp import (
    BaseRepository,
    BaseUnitOfWork,
    CrudOperation,
    Entity,
    FieldTypeSet,
    Model,
    exc,
)
from gen_epix.fastapp.repositories.dict.unit_of_work import DictUnitOfWork
from gen_epix.filter import BooleanOperator, CompositeFilter, Filter


class DictRepository(BaseRepository):
    @staticmethod
    def from_pkl(
        repository_class: Type[BaseRepository],
        entities: Iterable[Entity],
        file: str,
        **kwargs: dict,
    ) -> "DictRepository":
        if file.lower().endswith(".gz"):
            with gzip.open(file, "rb") as handle:
                db = pickle.load(handle)
        else:
            with open(file, "rb") as handle:
                db = pickle.load(handle)
        # TODO: check validity of db
        repository = repository_class(entities, db, **kwargs)
        assert isinstance(repository, DictRepository)
        return repository

    def __init__(
        self,
        entities: Iterable[Entity],
        db: dict[Type[Model], dict[Hashable, Model]],
        **kwargs: dict,
    ):
        extra_data = kwargs.pop("extra_data", "ignore")
        missing_data = kwargs.pop("missing_data", "raise")
        timestamp_factory = kwargs.pop("timestamp_factory", datetime.datetime.now)
        if extra_data not in {"ignore", "raise", "drop"}:
            raise ValueError(f"Invalid extra_data: {extra_data}")
        if missing_data not in {"raise", "ignore"}:
            raise ValueError(f"Invalid missing_data: {missing_data}")
        # Initialize properties
        self._db = dict(db.items())
        self._timestamp_factory = timestamp_factory
        self._entities = set(entities)
        self._links = {}
        self._get_id: dict[Type[Model], Callable[[Model], Hashable]] = {}
        self._back_links: dict[Type[Model], list[tuple[Type[Model], str]]] = {}
        self._value_field_names: dict[type[Model], list[str]] = {}
        self._keys_generators: dict[type[Model], dict[int, Callable[[Model], str]]] = {}
        self._init_properties(entities, db, missing_data)

        self._verify_extra_models_and_extract_reverse_links(extra_data)

    def _init_properties(
        self, entities: Iterable[Entity], db: dict, missing_data: str
    ) -> None:
        # Further populate properties
        for entity in entities:
            # Extract entity data
            model_class = entity.model_class
            assert issubclass(model_class, Model)
            self._links[model_class] = self._get_links(entity)
            self._back_links[model_class] = []
            self._value_field_names[model_class] = []
            for field_type in FieldTypeSet.DATA.value:
                self._value_field_names[model_class].extend(
                    entity.get_field_names(field_type=field_type)
                )
            self._keys_generators[model_class] = entity.get_keys_generator()
            if entity.persistable and model_class not in db:
                if missing_data == "ignore":
                    self._db[model_class] = {}
                elif missing_data == "raise":
                    raise ValueError(f"No data for model {model_class}")
                else:
                    raise NotImplementedError
            id_field_name = entity.id_field_name
            if not isinstance(id_field_name, str):
                raise NotImplementedError(
                    f"Model {model_class.__name__} has more than one ID field"
                )
            self._get_id[model_class] = lambda x: getattr(x, id_field_name)

    def _verify_extra_models_and_extract_reverse_links(self, extra_data: str) -> None:
        # Verify extra Models in db and extract reverse links
        for model_class, links in self._links.items():
            to_pop = []
            for i, link in enumerate(links):
                link_field_name, link_model_class, _, _, _ = link
                if link_model_class not in self._links:
                    if extra_data == "ignore":
                        continue
                    if extra_data == "drop":
                        to_pop.append(i)
                        continue
                    if extra_data == "raise":
                        raise ValueError(
                            f"Model {model_class} links to "
                            "additional linked model {link_model_class}"
                        )
                    raise NotImplementedError
                self._back_links[link_model_class].append(
                    (model_class, link_field_name)
                )
            for i in reversed(to_pop):
                links.pop(i)

    def crud(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable | None,
        model_class: Type[Model],
        objs: Model | Iterable[Model] | None,
        obj_ids: Hashable | Iterable[Hashable] | None,
        operation: CrudOperation,
        filter: Filter | None = None,
        **kwargs: dict,
    ) -> Hashable | list[Hashable] | Model | list[Model] | bool | list[bool] | None:
        BaseRepository.verify_crud_args(model_class, objs, obj_ids, operation)
        match operation:
            case CrudOperation.READ_ALL:
                return self.read_all(model_class, filter, **kwargs)
            case CrudOperation.READ_ONE:
                return self.read_one(model_class, obj_ids, **kwargs)
            case CrudOperation.READ_SOME:
                return self.read_some(model_class, obj_ids, **kwargs)
            case CrudOperation.CREATE_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=True,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.CREATE_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=True,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.UPDATE_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=False,
                    raise_on_missing=True,
                    **kwargs,
                )
            case CrudOperation.UPDATE_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=False,
                    raise_on_missing=True,
                    **kwargs,
                )
            case CrudOperation.UPSERT_ONE:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=False,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.UPSERT_SOME:
                return self.upsert_some(
                    user_id,
                    model_class,
                    objs,
                    raise_on_present=False,
                    raise_on_missing=False,
                    **kwargs,
                )
            case CrudOperation.DELETE_ONE:
                return self.delete_some(model_class, obj_ids, **kwargs)
            case CrudOperation.DELETE_SOME:
                return self.delete_some(model_class, obj_ids, **kwargs)
            case CrudOperation.DELETE_ALL:
                return self.delete_all(model_class, filter, **kwargs)
            case CrudOperation.EXISTS_ONE:
                return self.exists_one(model_class, obj_ids)
            case CrudOperation.EXISTS_SOME:
                return self.exists_some(model_class, obj_ids)
            case _:
                raise NotImplementedError(f"Operation {operation} not implemented")

    def uow(self, **kwargs: dict) -> BaseUnitOfWork:
        return DictUnitOfWork()

    def split_filter(
        self, model_class: Type, filter: Filter | None
    ) -> tuple[Filter | None, Filter | None]:
        # Entire filter can be used as where clause since the data are stored as
        # domain models
        return filter, None

    def verify_valid_ids(
        self,
        uow: BaseUnitOfWork,
        user_id: Hashable,
        model_class: Type[Model],
        obj_ids: Iterable[Hashable],
        verify_exists: bool = True,
        verify_duplicate: bool = True,
    ) -> None:
        if verify_exists:
            df = self._db[model_class]
            invalid_obj_ids = [x for x in obj_ids if x not in df]
            if invalid_obj_ids:
                DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)
        if verify_duplicate:
            DictRepository._verify_duplicate_ids(model_class, obj_ids)

    def read_all(
        self,
        model_class: Type[Model],
        filter: Filter | None,
        cascade_read: bool = False,
        return_id: bool = False,
        **kwargs: dict,
    ) -> list[Model]:
        df = self._db[model_class]
        # Get any query filter
        obj_filter = kwargs.get("obj_filter")
        if filter and obj_filter:
            query_filter = CompositeFilter(
                filters=[filter, obj_filter], operator=BooleanOperator.AND
            )
        elif filter:
            query_filter = filter
        elif obj_filter:
            query_filter = obj_filter
        else:
            query_filter = None
        # Get matching objects
        if query_filter:
            if return_id:
                objs = [
                    x
                    for x, y in zip(
                        df.keys(), query_filter.match_rows(df.values(), is_model=True)
                    )
                    if y
                ]
            else:
                objs = list(query_filter.filter_rows(df.values(), is_model=True))
        elif return_id:
            objs: list[Hashable] = list(df.keys())
        else:
            objs: list[Model] = list(df.values())
        # Make copy of objects for returning
        if not return_id:
            objs = [x.model_copy() for x in objs if x]
        # Cascade read linked objects if necessary
        if cascade_read and not return_id:
            self._cascade_read(model_class, objs)
        return objs

    def read_one(
        self,
        model_class: Type[Model],
        obj_id: Hashable,
        cascade_read: bool = False,
        return_id: bool = False,
        allow_duplicate_ids: bool = False,
        **kwargs: dict,
    ) -> Model:
        return self.read_some(
            model_class,
            [obj_id],
            cascade_read=cascade_read,
            return_id=return_id,
            allow_duplicate_ids=allow_duplicate_ids,
            **kwargs,
        )[0]

    def read_some(
        self,
        model_class: Type[Model],
        obj_ids: Iterable[Hashable],
        cascade_read: bool = False,
        return_id: bool = False,
        allow_duplicate_ids: bool = False,
        **kwargs: dict,
    ) -> list[Model]:
        df = self._db[model_class]
        # Read some or one
        if return_id:
            objs: list[Hashable] = list(obj_ids)
            invalid_obj_ids = [x for x in objs if x not in df]
            if invalid_obj_ids:
                DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)
            if not allow_duplicate_ids:
                DictRepository._verify_duplicate_ids(model_class, obj_ids)
        else:
            objs: list[Model] = [df.get(x) for x in obj_ids]
            # Verify input
            DictRepository._verify_valid_ids(model_class, obj_ids, objs)
            if not allow_duplicate_ids:
                DictRepository._verify_duplicate_ids(model_class, obj_ids)

        # Make copy of objects for returning
        if not return_id:
            objs = [x.model_copy() for x in objs if x]

        # Cascade read linked objects if necessary
        if cascade_read and not return_id:
            self._cascade_read(model_class, objs)
        return objs

    def upsert_some(
        self,
        user_id: Hashable,
        model_class: Type[Model],
        objs: Model | Iterable[Model],
        raise_on_present: bool = False,
        raise_on_missing: bool = False,
        return_id: bool = False,
        **kwargs: dict,
    ) -> Hashable | list[Hashable] | Model | list[Model]:
        df = self._db[model_class]
        get_id = self._get_id[model_class]
        is_iterable, objs = DictRepository._to_iterable(objs)
        obj_ids: list[Hashable] = [get_id(x) for x in objs]  # type: ignore
        df_objs: list[Model | None] = [df.get(x) for x in obj_ids]
        # Verify input
        DictRepository._verify_duplicate_ids(model_class, obj_ids)
        invalid_type_ids = [get_id(x) for x in objs if not isinstance(x, model_class)]
        self._verify_upsert_objects(
            invalid_type_ids,
            model_class,
            obj_ids,
            df_objs,
            raise_on_present,
            raise_on_missing,
        )
        DictRepository._verify_duplicate_keys(
            get_id, self._keys_generators[model_class], model_class, objs, df.values()
        )

        # Upsert objects
        value_field_names = self._value_field_names[model_class]
        links = self._links[model_class]
        links = [tuple([x[0], x[2], x[4]]) for x in links]
        current_time = self._timestamp_factory()
        for i, obj, df_obj in zip(range(len(df_objs)), objs, df_objs):
            if df_obj:
                # Already existing -> update df_obj with obj data
                for field_name in value_field_names:
                    # Update value field
                    setattr(df_obj, field_name, getattr(obj, field_name))
                for link_field_name, relationship_field_name, linked_df in links:
                    # Verify and update link
                    linked_obj_id = getattr(obj, link_field_name)
                    if not linked_obj_id:
                        setattr(df_obj, link_field_name, None)
                        setattr(df_obj, relationship_field_name, None)
                        continue
                    if linked_df is not None and linked_obj_id not in linked_df:
                        raise exc.InvalidIdsError(
                            (
                                f"Model {model_class}: obj {get_id(obj)} has invalid id"
                                f' in {link_field_name}: "{linked_obj_id}"'
                            ),
                            ids=[linked_obj_id],
                        )
                    setattr(df_obj, link_field_name, linked_obj_id)
                    linked_obj = getattr(obj, relationship_field_name)
                    if not linked_obj:
                        continue
                    get_link_id = self._get_id[linked_obj.__class__]
                    if get_link_id(linked_obj) != linked_obj_id:
                        raise exc.InvalidLinkIdsError(
                            (
                                f"Model {model_class}: obj {get_link_id(obj)} has different id "
                                f'in {link_field_name} ("{linked_obj_id}") versus '
                                f'{relationship_field_name} ("{get_link_id(linked_obj)}")'
                            ),
                            ids=[linked_obj_id, get_link_id(linked_obj)],
                        )
            else:
                # New -> insert copy of obj
                df_obj: Model = obj.model_copy()  # type: ignore
                df[get_id(df_obj)] = df_obj  # type: ignore
                df_objs[i] = df_obj
        if return_id:
            return obj_ids if is_iterable else obj_ids[0]
        return (
            [x.model_copy() for x in df_objs]
            if is_iterable
            else df_objs[0].model_copy()
        )

    def delete_some(
        self,
        model_class: Type[Model],
        obj_ids: Hashable | Iterable[Hashable],
        **kwargs: dict,
    ) -> Hashable | list[Hashable] | None:
        df = self._db[model_class]
        is_iterable, obj_ids = DictRepository._to_iterable(obj_ids)
        df_objs = [df.get(x) for x in obj_ids]
        back_links = self._back_links[model_class]

        # Verify input
        DictRepository._verify_valid_ids(model_class, obj_ids, df_objs)
        DictRepository._verify_duplicate_ids(model_class, obj_ids)

        # Verify existence of link (foreign key) constraint conflicts
        uq_obj_ids = set(obj_ids)
        for link_model_class, link_field_name in back_links:
            link_df = self._db[link_model_class]
            get_id = self._get_id[link_model_class]
            linked_obj_ids = [
                get_id(x)
                for x in link_df.values()
                if getattr(x, link_field_name) in uq_obj_ids
            ]
            if linked_obj_ids:
                linked_obj_ids_str = ", ".join([f'"{x}"' for x in linked_obj_ids])
                raise exc.LinkConstraintViolationError(
                    (
                        f"Model {model_class}: link constraint conflict in model "
                        f"{link_model_class}, id(s): {linked_obj_ids_str}"
                    ),
                    list(uq_obj_ids),
                    linked_obj_ids,
                )
        # Delete objects
        for obj_id in uq_obj_ids:
            df.pop(obj_id)
        return obj_ids if is_iterable else obj_ids[0]

    def delete_all(
        self, model_class: Type[Model], filter: Filter | None, **kwargs: dict
    ) -> list[Hashable]:
        df = self._db[model_class]
        # Get any query filter
        obj_filter = kwargs.get("obj_filter")
        if filter and obj_filter:
            query_filter = CompositeFilter(
                filters=[filter, obj_filter], operator=BooleanOperator.AND
            )
        elif filter:
            query_filter = filter
        elif obj_filter:
            query_filter = obj_filter
        else:
            query_filter = None
        # Delete objects
        if query_filter:
            # Delete objects matching the query filter
            obj_ids = [
                x
                for x, y in zip(
                    df.keys(), query_filter.match_rows(df.values(), is_model=True)
                )
                if y
            ]
            for obj_id in obj_ids:
                self._db[model_class].pop(obj_id)
        else:
            # Delete all objects
            obj_ids = list(self._db[model_class].keys())
            self._db[model_class] = {}
        return obj_ids

    def exists_one(self, model_class: Type[Model], obj_id: Hashable) -> bool:
        df = self._db[model_class]
        return obj_id in df

    def exists_some(
        self, model_class: Type[Model], obj_ids: Iterable[Hashable]
    ) -> list[bool]:
        df = self._db[model_class]
        return [x in df for x in obj_ids]

    def _get_links(
        self, entity: Entity
    ) -> list[tuple[str, Type[Model], str, int, dict[Hashable, Model] | None]]:
        # Return list[tuple[link_field_name, LinkModel, relationship_field_name, link_type_id, linked_df|None]]
        links = []
        for link_field_name in entity.get_link_field_names():
            (
                link_type_id,
                link_model_class,
                relationship_field_name,
            ) = entity.get_link_properties_by_field_name(link_field_name)
            links.append(
                (
                    link_field_name,
                    link_model_class,
                    relationship_field_name,
                    link_type_id,
                    self._db.get(link_model_class),
                )
            )
        return links

    def _cascade_read(
        self,
        model_class: Type[Model],
        objs: list[Model],
    ) -> None:
        for (
            link_field_name,
            link_model_class,
            relationship_field_name,
            _,
            linked_df,
        ) in self._links[model_class]:
            if linked_df is None:
                continue
            linked_obj_ids = [
                getattr(x, link_field_name) for x in objs if getattr(x, link_field_name)
            ]
            linked_objs = self.read_some(
                link_model_class,
                linked_obj_ids,
                cascade_read=False,
                allow_duplicate_ids=True,
            )
            for obj, linked_obj in zip(objs, linked_objs):
                setattr(obj, relationship_field_name, linked_obj)

    def _verify_upsert_objects(
        self,
        invalid_type_ids: list[Hashable],
        model_class: Type[Model],
        obj_ids: list[Hashable],
        df_objs: list[Model | None],
        raise_on_present: bool,
        raise_on_missing: bool,
    ) -> None:
        if invalid_type_ids:
            invalid_type_ids_str = ", ".join([f'"{x}"' for x in invalid_type_ids])
            raise exc.InvalidModelIdsError(
                f"Model {model_class}: object(s) are of different type: {invalid_type_ids_str}",
                ids=invalid_type_ids,
            )
        get_id = self._get_id[model_class]
        if raise_on_present:
            present_ids = [get_id(x) for x in df_objs if x is not None]
            if present_ids:
                present_ids_str = ", ".join(['"{x}"' for x in present_ids])
                raise exc.AlreadyExistingIdsError(
                    f"{model_class} object(s) already exist: {present_ids_str}",
                    ids=present_ids,
                )
        if raise_on_missing:
            missing_ids = [x for x, y in zip(obj_ids, df_objs) if y is None]
            if missing_ids:
                missing_ids_str = ", ".join([f"{x}" for x in missing_ids])
                raise exc.InvalidIdsError(
                    f"{model_class} object(s) do not exist: {missing_ids_str}",
                    ids=missing_ids,
                )

    @staticmethod
    def _verify_valid_ids(
        model_class: Type[Model],
        obj_ids: Iterable[Hashable],
        objs: Iterable[Model | None],
    ) -> None:
        invalid_obj_ids = [x for x, y in zip(obj_ids, objs) if y is None]
        if invalid_obj_ids:
            DictRepository._raise_invalid_ids(model_class, invalid_obj_ids)

    @staticmethod
    def _raise_invalid_ids(
        model_class: Type[Model], invalid_obj_ids: Iterable[Hashable]
    ) -> None:
        invalid_obj_ids_str = ", ".join(f'"{x}"' for x in invalid_obj_ids)
        raise exc.InvalidIdsError(
            f"Model {model_class}: invalid object id(s) provided: {invalid_obj_ids_str}",
            ids=invalid_obj_ids,
        )

    @staticmethod
    def _verify_duplicate_ids(
        model_class: Type[Model], obj_ids: Iterable[Hashable]
    ) -> None:
        set_ = set()
        duplicate_ids = [x for x in obj_ids if x in set_ or set_.add(x)]
        if duplicate_ids:
            DictRepository._raise_duplicate_ids(model_class, duplicate_ids)

    @staticmethod
    def _raise_duplicate_ids(
        model_class: Type[Model], duplicate_ids: Iterable[Hashable]
    ) -> None:
        duplicate_ids_str = ", ".join([f'"{x}"' for x in duplicate_ids])
        raise exc.DuplicateIdsError(
            f"Model {model_class}: object ids are not unique: {duplicate_ids_str}",
            ids=duplicate_ids,
        )

    @staticmethod
    def _verify_duplicate_keys(
        get_id: Callable[[Model], Hashable],
        keys_generator: Callable,
        model_class: Type[Model],
        objs: list[Model],
        df_objs: list[Model] | None,
    ) -> None:
        if not objs:
            # No objs -> no duplicates
            return
        keys = keys_generator(objs[0])
        if not keys:
            # No keys -> no duplicates
            return
        key_ids = list(keys.keys())

        def get_keys(obj):
            keys = keys_generator(obj)
            return tuple(keys[x] for x in key_ids)

        # Check for duplicate keys among objs
        obj_keys_list = [get_keys(x) for x in objs]
        n_keys = len(keys)
        duplicate_objs = []
        for i in range(n_keys):
            curr_obj_keys_list = [x[i] for x in obj_keys_list]
            curr_obj_keys = set(curr_obj_keys_list)
            if len(curr_obj_keys) < len(curr_obj_keys_list):
                seen = set()
                uq_obj_keys = set(
                    x for x in curr_obj_keys_list if x not in seen and not seen.add(x)  # type: ignore
                )
                duplicate_obj_keys = curr_obj_keys - uq_obj_keys
                duplicate_objs += [
                    x
                    for x, y in zip(objs, curr_obj_keys_list)
                    if y in duplicate_obj_keys
                ]
        if duplicate_objs:
            raise exc.UniqueConstraintViolationError(
                f"Model {model_class}: object keys are not unique",
                duplicate_key_ids=list(set([get_id(x) for x in duplicate_objs])),
            )
        # Check for duplicate keys between objs and df_objs, excluding those df_objs
        #  that have the same id as an obj
        if not df_objs:
            return
        obj_ids = set([get_id(x) for x in objs])
        df_obj_keys = [get_keys(x) for x in df_objs if get_id(x) not in obj_ids]
        duplicate_objs = []
        for i in range(n_keys):
            curr_df_obj_keys = set(x[i] for x in df_obj_keys)
            curr_obj_keys_list = [x[i] for x in obj_keys_list]
            curr_obj_keys = set(curr_obj_keys_list)
            duplicate_obj_keys = curr_obj_keys & curr_df_obj_keys
            if duplicate_obj_keys:
                duplicate_objs += [
                    x
                    for x, y in zip(objs, curr_obj_keys_list)
                    if y in duplicate_obj_keys
                ]
        if duplicate_objs:
            raise exc.UniqueConstraintViolationError(
                f"Model {model_class}: object keys are not unique",
                duplicate_key_ids=list(set([get_id(x) for x in duplicate_objs])),
            )

    @staticmethod
    def _to_iterable(
        obj: Any | Iterable[Any],
    ) -> tuple[bool, list[Any]]:
        # TODO: take model_class as argument, and derive id field class from it to use for check
        if isinstance(obj, (list, set, frozenset)):
            return True, obj
        if issubclass(type(obj), Model):
            return False, [obj]
        if isinstance(obj, UUID):
            return False, [obj]
        try:
            list_obj = list(obj)
            if not list_obj:
                return True, list_obj
            if issubclass(type(list_obj[0]), Model):
                return False, list_obj
            return False, [obj]
        except:
            return False, [obj]



import abc

from gen_epix.fastapp import BaseService
from gen_epix.seqdb.domain.enum import ServiceType


class BaseAbacService(BaseService):
    SERVICE_TYPE = ServiceType.ABAC

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)

    @abc.abstractmethod
    def register_policies(self) -> None:
        raise NotImplementedError



from typing import Callable
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator
from gen_epix.omopdb.api.base import EXCLUDED_PERMISSIONS
from gen_epix.omopdb.domain import command, enum, model
from gen_epix.omopdb.domain.model import CompleteUser


class UpdateUserRequestBody(PydanticBaseModel):
    is_active: bool | None
    organization_id: UUID | None


class UpdateUserOwnOrganizationRequestBody(PydanticBaseModel):
    organization_id: UUID


def create_organization_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    **kwargs: dict,
) -> None:

    assert handle_exception

    @router.get(
        "/user_me",
        operation_id="user_me__get_one",
        name="UserMe",
    )
    async def user_me__get_one(
        user: registered_user_dependency,  # type: ignore
    ) -> CompleteUser:
        try:
            cmd = command.RetrieveCompleteUserCommand(user=user)
            app_retval: model.CompleteUser = app.handle(cmd)
            retval = CompleteUser.from_model(app_retval)
        except Exception as exception:
            handle_exception("f98b34ec", user, exception)
        return retval

    @router.put(
        "/users/{object_id}",
        operation_id="users__put_one",
        name="UpdateUser",
    )
    async def users__put_one(
        user: registered_user_dependency, object_id: UUID, request_body: UpdateUserRequestBody  # type: ignore
    ) -> model.User:
        try:
            cmd = command.UpdateUserCommand(
                user=user,
                tgt_user_id=object_id,
                is_active=request_body.is_active,
                organization_id=request_body.organization_id,
            )
            retval: model.User = app.handle(cmd)
        except Exception as exception:
            handle_exception("a594ba2b", None, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.ORGANIZATION,
        user_dependency=registered_user_dependency,
        excluded_permissions=EXCLUDED_PERMISSIONS,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



from datetime import timedelta

from cachetools import TTLCache, cached

from gen_epix.fastapp import CrudOperation, EventTiming
from gen_epix.omopdb.domain import command, exc, model
from gen_epix.omopdb.domain.repository.organization import BaseOrganizationRepository
from gen_epix.omopdb.domain.service.organization import BaseOrganizationService
from gen_epix.omopdb.policies.update_user_policy import UpdateUserPolicy


class OrganizationService(BaseOrganizationService):
    DEFAULT_CFG = {
        "user_invitation_time_to_live": 86400,  # 1 day
    }

    def register_policies(self) -> None:
        f = self.app.register_policy
        policy = UpdateUserPolicy(self)
        for command_class in BaseOrganizationService.UPDATE_USER_COMMANDS:
            f(command_class, policy, EventTiming.BEFORE)

    @cached(cache=TTLCache(maxsize=1000, ttl=60))
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        with self.repository.uow() as uow:
            return self.repository.retrieve_user_by_key(uow, user_key)

    def invite_user(
        self,
        cmd: command.InviteUserCommand,
    ) -> model.UserInvitation:
        user = cmd.user
        if user is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        if user.id is None:
            raise exc.UnauthorizedAuthError("User has no ID")
        email = cmd.email
        initial_roles = cmd.roles
        initial_data_collection_ids = cmd.data_collection_ids

        with self.repository.uow() as uow:
            # Verify if data_collection_ids are valid
            self.repository.crud(
                uow,
                user.id,
                model.DataCollection,
                None,
                initial_data_collection_ids,
                CrudOperation.READ_SOME,
            )
            # Verify if user already exists
            try:
                self.app.user_manager.retrieve_user_by_key(user)
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "acba1a0e",
                            f"User {email} already exists",
                        )
                    )
                raise exc.UserAlreadyExistsAuthError("User already exists")
            except exc.UserAlreadyExistsAuthError as e:
                raise e
            except Exception:
                # User does not exist yet
                pass
            # Verify if invitation(s) already exists for this email, and delete those
            # TODO: Must be done within the same session to be safe,
            # so requires specific repository method
            user_invitations: list[model.UserInvitation] = self.repository.crud(
                uow,
                user.id,
                model.UserInvitation,
                None,
                None,
                CrudOperation.READ_ALL,
            )  # type: ignore
            user_invitations = [x for x in user_invitations if x.email == email]
            if user_invitations:
                self.repository.crud(
                    uow,
                    user.id,
                    model.UserInvitation,
                    None,
                    [x.id for x in user_invitations],  # type: ignore
                    CrudOperation.DELETE_SOME,  # Invitations should not be undeletable
                )  # type: ignore
            # Create user invitation
            user_invitation = model.UserInvitation(
                email=email,
                roles=initial_roles,
                data_collection_ids=initial_data_collection_ids,
                invited_by_user_id=user.id,
                token=self.generate_user_invitation_token(),
                expires_at=self.generate_timestamp()
                + timedelta(
                    seconds=self.props.get(
                        "user_invitation_time_to_live",
                        OrganizationService.DEFAULT_CFG["user_invitation_time_to_live"],
                    )
                ),
            )
            user_invitation_in_db: model.UserInvitation = self.app.handle(
                command.UserInvitationCrudCommand(
                    user=user,
                    objs=user_invitation,
                    operation=CrudOperation.CREATE_ONE,
                )
            )
        return user_invitation_in_db

    def register_invited_user(
        self, cmd: command.RegisterInvitedUserCommand
    ) -> model.User:
        new_user = cmd.user
        if new_user is None:
            # Should not happen
            raise AssertionError("Command has no user")
        if not self.app.user_manager:
            raise exc.InvalidArgumentsError("User manager not set")

        assert isinstance(self.repository, BaseOrganizationRepository)
        with self.repository.uow() as uow:
            # Get possible user invitations
            user_invitations: list[model.UserInvitation] = self.repository.crud(  # type: ignore
                uow,
                None,
                model.UserInvitation,
                None,
                None,
                CrudOperation.READ_ALL,
            )
            now = self.generate_timestamp()
            user_invitations = [
                x
                for x in user_invitations
                if x.email == new_user.email and x.expires_at > now
            ]
            if not user_invitations:
                raise exc.UnauthorizedAuthError(
                    f"No valid invitations found for user {new_user.email}",
                )
            user_invitations_with_token = [
                x for x in user_invitations if x.token == cmd.token
            ]
            if not user_invitations_with_token:
                raise exc.UnauthorizedAuthError(
                    f"No invitation found for token {cmd.token}"
                )
            # Choose the invitation with the latest expiry date
            user_invitation: model.UserInvitation = sorted(
                user_invitations_with_token, key=lambda x: x.expires_at
            )[-1]
            # Create user
            user_in_db = self.app.user_manager.create_new_user_from_token(
                new_user,
                user_invitation.token,
                created_by_user_id=user_invitation.invited_by_user_id,
                roles=user_invitation.roles,
            )
        return user_in_db  # type: ignore

    def retrieve_complete_user(
        self, cmd: command.RetrieveCompleteUserCommand
    ) -> model.CompleteUser:
        user = cmd.user
        if user is None:
            raise exc.UnauthorizedAuthError("Command has no user")

        # Get data collections
        data_collections: list[model.DataCollection] = self.repository.crud(
            cmd.user,
            model.DataCollection,
            None,
            user.data_collection_ids,
            CrudOperation.READ_SOME,
        )

        return model.CompleteUser(
            **{x: y for x, y in user.model_dump().items() if x != "data_collections"},
            data_collections=data_collections,
        )

    def update_user(
        self,
        cmd: command.UpdateUserCommand,
    ) -> model.User:
        tgt_user = self.repository.crud(
            cmd.user,
            model.User,
            None,
            cmd.tgt_user_id,
            CrudOperation.READ_ONE,
        )
        is_active = tgt_user.is_active if cmd.is_active is None else cmd.is_active
        roles = tgt_user.roles if cmd.roles is None else cmd.roles
        data_collection_ids = (
            tgt_user.data_collection_ids
            if cmd.data_collection_ids is None
            else cmd.data_collection_ids
        )
        # Special case: no updates
        if (
            tgt_user.is_active == is_active
            and tgt_user.roles == roles
            and tgt_user.data_collection_ids == data_collection_ids
        ):
            return tgt_user
        # Check if data_collection_ids exists
        if tgt_user.data_collection_ids != data_collection_ids:
            self.repository.crud(
                cmd.user,
                model.DataCollection,
                None,
                data_collection_ids,
                CrudOperation.READ_SOME,
            )
        # Update user
        tgt_user.is_active = is_active
        tgt_user.roles = roles
        tgt_user.data_collection_ids = data_collection_ids
        return self.repository.crud(
            cmd.user,
            model.User,
            tgt_user,
            None,
            CrudOperation.UPDATE_ONE,
        )



from gen_epix.fastapp.repositories import DictRepository
from gen_epix.omopdb.domain.repository.system import BaseSystemRepository


class SystemDictRepository(DictRepository, BaseSystemRepository):
    pass



from gen_epix.omopdb.services.omop import OmopService as OmopService
from gen_epix.omopdb.services.organization import (
    OrganizationService as OrganizationService,
)
from gen_epix.omopdb.services.rbac import RbacService as RbacService
from gen_epix.omopdb.services.system import SystemService as SystemService
from gen_epix.omopdb.services.user_manager import UserManager as UserManager



from gen_epix.fastapp import Command, CrudOperation
from gen_epix.omopdb.domain import command, enum, model
from gen_epix.omopdb.domain.policy.organization import BaseUpdateUserPolicy


class UpdateUserPolicy(BaseUpdateUserPolicy):
    def is_allowed(self, cmd: Command) -> bool:
        if cmd.user is None:
            return False

        # Root user can do anything
        if enum.Role.ROOT in cmd.user.roles:
            return True

        # Check if user is allowed to update themselves
        if isinstance(cmd, command.UpdateUserCommand):
            if cmd.tgt_user_id == cmd.user.id:
                # Only root user can update themselves
                return enum.Role.ROOT in cmd.user.roles
        elif isinstance(cmd, command.InviteUserCommand):
            if cmd.user.email == cmd.email:
                # User cannot invite themselves
                return False
        else:
            raise NotImplementedError()

        # Check if invite is allowed according to roles
        if isinstance(cmd, command.InviteUserCommand):
            # Only admin can invite another, except root and admin
            if enum.Role.ADMIN not in cmd.user.roles:
                return False
            if enum.Role.ROOT in cmd.roles or enum.Role.ADMIN in cmd.roles:
                return False

        # Check if update is allowed according to roles
        if isinstance(cmd, command.UpdateUserCommand):
            # Only admin can update another, except root
            if enum.Role.ADMIN not in cmd.user.roles:
                return False
            with self.organization_service.repository.uow() as uow:
                tgt_user: model.User = self.organization_service.repository.crud(
                    uow,
                    cmd.user.id,
                    model.User,
                    None,
                    cmd.tgt_user_id,
                    CrudOperation.READ_ONE,
                )
            if enum.Role.ROOT in tgt_user.roles:
                return False

        raise NotImplementedError()



from uuid import UUID

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from gen_epix.fastapp.enum import AuthProtocol, OauthFlowType


class IDPUser(PydanticBaseModel):
    issuer: str = Field(description="The issuer of the user")
    sub: str = Field(description="The sub of the user")


class IdentityProvider(PydanticBaseModel):
    name: str = Field(description="Name of the identity provider")
    label: str = Field(description="Label of the identity provider")
    issuer: str = Field(description="The issuer URL of the identity provider")
    auth_protocol: AuthProtocol = Field(description="The authentication protocol")
    oauth_flow: OauthFlowType | None = Field(
        default=None, description="The OAuth flow type of the identity provider"
    )
    discovery_url: str | None = Field(
        default=None, description="The discovery URL of the identity provider"
    )
    client_id: str | None = Field(
        default=None, description="The client ID of the identity provider"
    )
    client_secret: str | None = Field(
        default=None, description="The client secret of the identity provider"
    )
    scope: str | None = Field(
        default=None, description="The scope of the identity provider"
    )


class Claims(PydanticBaseModel):
    scheme: str = Field(description="The authorization scheme of the token")
    token: str = Field(description="The original token containing the claims")
    idp_client_id: UUID = Field(
        description="The ID of the IDP client that processed the claims"
    )
    claims: dict[str, str | int | bool | list[str]] = Field(
        description="The claims as verified and processed by the IDP client"
    )


class OIDCConfiguration(PydanticBaseModel):
    name: str = Field(description="Service name")
    label: str = Field(description="Service label")
    discovery_url: str = Field(
        description="The URL of the OpenID Connect discovery document"
    )
    client_id: str = Field(description="The client ID of the application")
    client_secret: str | None = Field(
        default=None, description="The client secret of the application"
    )
    scope: str = Field(description="The scope of the application")



from gen_epix.omopdb.repositories.omop_dict import (
    OmopDictRepository as OmopDictRepository,
)
from gen_epix.omopdb.repositories.omop_sa import OmopSARepository as OmopSARepository
from gen_epix.omopdb.repositories.organization_dict import (
    OrganizationDictRepository as OrganizationDictRepository,
)
from gen_epix.omopdb.repositories.organization_sa import (
    OrganizationSARepository as OrganizationSARepository,
)
from gen_epix.omopdb.repositories.system_dict import (
    SystemDictRepository as SystemDictRepository,
)
from gen_epix.omopdb.repositories.system_sa import (
    SystemSARepository as SystemSARepository,
)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import datetime
import json
from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer, field_validator

from gen_epix.casedb.domain import DOMAIN, enum
from gen_epix.casedb.domain.model.base import Model
from gen_epix.casedb.domain.model.geo import Region
from gen_epix.fastapp import User as ServiceUser
from gen_epix.fastapp.domain import Entity, create_keys, create_links

_SERVICE_TYPE = enum.ServiceType.ORGANIZATION
_ENTITY_KWARGS = {
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class Organization(Model):
    """
    Represents an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organizations",
        table_name="organization",
        persistable=True,
        keys=create_keys({1: "name", 2: "legal_entity_code"}),
        links=create_links(
            {
                1: ("legal_region_id", Region, "legal_region"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    name: str = Field(
        description="The name of the organization, UNIQUE", max_length=255
    )
    legal_entity_code: str = Field(
        description="The legal entity code of the organization, UNIQUE", max_length=255
    )
    legal_region_id: UUID | None = Field(
        default=None,
        description=(
            "The ID of the region that "
            "the organization is legally responsible for. FOREIGN KEY"
        ),
    )
    legal_region: Region | None = Field(
        default=None, description="The region corresponding to the ID"
    )


class UserNameEmail(ServiceUser, Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_name_emails",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    id: UUID | None = Field(default=None, description="The ID of the user")
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )
    email: str = Field(description="The email of the user", max_length=320)


class User(ServiceUser, Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="users",
        table_name="user",
        persistable=True,
        keys=create_keys({1: "email"}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    id: UUID | None = Field(default=None, description="The ID of the user")
    email: str = Field(description="The email of the user, UNIQUE", max_length=320)
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )

    is_active: bool = Field(
        default=True,
        description="Whether the user is active or not. An inactive user cannot perform any actions that require authorization.",
    )
    roles: set[enum.Role] = Field(description="The roles of the user", min_length=1)
    organization_id: UUID = Field(
        description="The ID of the organization of the user. FOREIGN KEY"
    )
    organization: Organization | None = Field(
        default=None, description="The organization of the user"
    )

    def get_key(self) -> str:
        return self.email

    @field_validator("roles", mode="before")
    @classmethod
    def _validate_roles(cls, value: set[enum.Role] | list[str] | str) -> set[enum.Role]:
        """
        Validate and convert roles representation to a set[Role]. When given as a
        string, it is assumed to be a JSON list of Role values.
        """
        if isinstance(value, str):
            return {enum.Role[x] for x in json.loads(value)}
        if isinstance(value, list):
            return {enum.Role[x] for x in value}
        return value

    @field_serializer("roles")
    def serialize_roles(self, value: set[enum.Role], _info):
        return [x.value for x in value]


class OrganizationSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_sets",
        table_name="organization_set",
        persistable=True,
        keys=create_keys({1: "name"}),
        **_ENTITY_KWARGS,
    )
    name: str = Field(description="The name of the organization set", max_length=255)
    description: str | None = Field(
        None, description="The description of the organization set."
    )


class OrganizationSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="organization_set_members",
        table_name="organization_set_member",
        persistable=True,
        keys=create_keys({1: ("organization_set_id", "organization_id")}),
        links=create_links(
            {
                1: ("organization_set_id", OrganizationSet, "organization_set"),
                2: ("organization_id", Organization, "organization"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    organization_set_id: UUID = Field(
        description="The ID of the organization set. FOREIGN KEY"
    )
    organization_set: OrganizationSet | None = Field(
        default=None, description="The organization set"
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: Organization | None = Field(None, description="The organization")


class Site(Model):
    """
    Represents a physical site of an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="sites",
        table_name="site",
        persistable=True,
        keys=create_keys({1: ("organization_id", "name")}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
                2: ("location_region_id", Region, "location_region"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    organization_id: UUID = Field(description="The ID of the organization. FOREIGN KEY")
    organization: Organization | None = Field(
        default=None, description="The organization corresponding to the ID"
    )
    name: str = Field(description="The name of an organization, UNIQUE", max_length=255)
    location_region_id: UUID = Field(
        description="The ID of the region within which the site is located. FOREIGN KEY"
    )
    location_region: Region | None = Field(
        default=None, description="The region corresponding to the ID"
    )


class Contact(Model):
    """
    A class representing contact information for an organization.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="contacts",
        table_name="contact",
        persistable=True,
        keys=create_keys({1: ("site_id", "name")}),
        links=create_links(
            {
                1: ("site_id", Site, "site"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    # TODO: Temporary implementation, check established models for this
    site_id: UUID | None = Field(
        description="The ID of the site in case the contact is site-specific. FOREIGN KEY",
    )
    site: Site | None = Field(
        default=None, description="The site corresponding to the ID"
    )
    name: str = Field(description="The name of the contact, UNIQUE", max_length=255)
    email: str | None = Field(
        default=None, description="The email address of the contact", max_length=320
    )
    phone: str | None = Field(
        default=None, description="The phone number of the contact"
    )


class IdentifierIssuer(Model):
    """
    A system or process that issues identifiers.
    The combination (identifier_issuer, issued_identifier) is universally unique.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="identifier_issuers",
        table_name="identifier_issuer",
        persistable=True,
        **_ENTITY_KWARGS,
    )
    name: str = Field(description="The name of the issuer", max_length=255)


class DataCollection(Model):
    """
    Represents a collection of data.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collections",
        table_name="data_collection",
        persistable=True,
        keys=create_keys({1: "name"}),
        **_ENTITY_KWARGS,
    )
    # TODO: Placeholder
    name: str = Field(
        description="The name of a data collection, UNIQUE", max_length=255
    )
    description: str | None = Field(
        default=None, description="The description of the data collection."
    )


class DataCollectionSet(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_sets",
        table_name="data_collection_set",
        persistable=True,
        keys=create_keys({1: "name"}),
        **_ENTITY_KWARGS,
    )
    name: str = Field(description="The name of the data collection set", max_length=255)
    description: str | None = Field(
        default=None, description="The description of the data collection set."
    )


class DataCollectionSetMember(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_set_members",
        table_name="data_collection_set_member",
        persistable=True,
        keys=create_keys(
            {1: ("data_collection_set_id", "data_collection_id")},
        ),
        links=create_links(
            {
                1: ("data_collection_set_id", DataCollectionSet, "data_collection_set"),
                2: ("data_collection_id", DataCollection, "data_collection"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    data_collection_set_id: UUID = Field(
        description="The ID of the data collection set. FOREIGN KEY"
    )
    data_collection_set: DataCollectionSet | None = Field(
        default=None, description="The data collection set"
    )
    data_collection_id: UUID = Field(
        description="The ID of the data collection. FOREIGN KEY"
    )
    data_collection: DataCollection | None = Field(
        default=None, description="The data collection"
    )


class DataCollectionRelation(Model):
    """
    Represents a directional relationship between two data collections.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="data_collection_relations",
        table_name="data_collection_relation",
        persistable=True,
        keys=create_keys({1: ("from_data_collection_id", "to_data_collection_id")}),
        links=create_links(
            {
                1: ("from_data_collection_id", DataCollection, "from_data_collection"),
                2: ("to_data_collection_id", DataCollection, "to_data_collection"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    from_data_collection_id: UUID = Field(
        description="The ID of the data collection the relation originates from. FOREIGN KEY"
    )
    from_data_collection: DataCollection | None = Field(
        default=None, description="The data collection the relation originates from"
    )
    to_data_collection_id: UUID = Field(
        description="The ID of the data collection the relation points to. FOREIGN KEY"
    )
    to_data_collection: DataCollection | None = Field(
        default=None, description="The data collection the relation points to"
    )
    share_case: bool = Field(
        description="Whether a case can be shared from one data collection to the other"
    )


class UserInvitation(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="user_invitations",
        table_name="user_invitation",
        persistable=True,
        keys=create_keys({1: ("email", "expires_at")}),
        links=create_links(
            {
                1: ("organization_id", Organization, "organization"),
                2: ("invited_by_user_id", User, "invited_by_user"),
            }
        ),
        **_ENTITY_KWARGS,
    )
    email: str = Field(description="The email of the user, UNIQUE", max_length=320)
    token: str = Field(description="The token of the invitation", max_length=255)
    expires_at: datetime.datetime = Field(
        description="The expiry date of the invitation"
    )
    roles: set[enum.Role] = Field(
        description="The identifiers of the initial roles of the user", min_length=1
    )
    invited_by_user_id: UUID = Field(
        description="The ID of the user who invited the user. FOREIGN KEY"
    )
    invited_by_user: User | None = Field(
        default=None, description="The user who invited the user"
    )
    organization_id: UUID = Field(
        description="The ID of the organization of the user. FOREIGN KEY"
    )
    organization: Organization | None = Field(
        default=None, description="The organization of the user"
    )

    @field_validator("roles", mode="before")
    @classmethod
    def _validate_roles(cls, value: set[enum.Role] | list[str] | str) -> set[enum.Role]:
        """
        Validate and convert roles representation to a set[Role]. When given as a
        string, it is assumed to be a JSON list of Role values.
        """
        if isinstance(value, str):
            return {enum.Role[x] for x in json.loads(value)}
        if isinstance(value, list):
            return {enum.Role[x] for x in value}
        return value

    @field_serializer("roles")
    def serialize_roles(self, value: set[enum.Role], _info):
        return [x.value for x in value]


# Non-persistable
class CaseShareGraph(Model):
    """
    Represents a graph of data collections that can share cases from on data collection
    to another. Edges corresponding to DataCollectionRelations that have
    share_case=True. Reflexive edges, corresponding to data collections to which new
    cases may be added, are included separately.
    """

    ENTITY: ClassVar = Entity(
        persistable=False,
        **_ENTITY_KWARGS,
    )
    forward: dict[UUID, set[UUID]] = Field(
        description="The directed graph of data collections that can share cases, as dict[from_data_collection_id: set[to_data_collection_id]]"
    )
    reverse: dict[UUID, set[UUID]] = Field(
        description="The directed graph of data collections that can share cases, as dict[from_data_collection_id: set[to_data_collection_id]]"
    )
    new_cases_allowed_data_collection_ids: set[UUID] = Field(
        description="The data collections to which new cases can be added"
    )


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



# pylint: disable=useless-import-alias

from typing import ClassVar, Self
from uuid import UUID

from pydantic import Field, field_validator, model_validator

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.casedb.domain.command.base import Command, UpdateAssociationCommand
from gen_epix.fastapp.services import auth
from gen_epix.filter import TypedDatetimeRangeFilter


# auth
class GetIdentityProvidersCommand(auth.GetIdentityProvidersCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.AUTH


# organization
class OrganizationSetOrganizationUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    ASSOCIATION_CLASS: ClassVar = model.OrganizationSetMember
    LINK_FIELD_NAME1: ClassVar = "organization_set_id"
    LINK_FIELD_NAME2: ClassVar = "organization_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.OrganizationSetMember]


class DataCollectionSetDataCollectionUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    ASSOCIATION_CLASS: ClassVar = model.DataCollectionSetMember
    LINK_FIELD_NAME1: ClassVar = "data_collection_set_id"
    LINK_FIELD_NAME2: ClassVar = "data_collection_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.DataCollectionSetMember]


class InviteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    email: str
    roles: set[enum.Role]
    organization_id: UUID


class RegisterInvitedUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    token: str


class RetrieveOrganizationContactCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    organization_ids: list[UUID] | None = None
    site_ids: list[UUID] | None = None
    contact_ids: list[UUID] | None = None


class UpdateUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    tgt_user_id: UUID
    is_active: bool | None
    roles: set[enum.Role] | None
    organization_id: UUID | None


# geo
class RetrieveContainingRegionCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.GEO

    region_ids: list[UUID]
    region_set_id: UUID
    level: int


# case


class CaseTypeSetCaseTypeUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE
    ASSOCIATION_CLASS: ClassVar = model.CaseTypeSetMember
    LINK_FIELD_NAME1: ClassVar = "case_type_set_id"
    LINK_FIELD_NAME2: ClassVar = "case_type_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.CaseTypeSetMember]


class CaseTypeColSetCaseTypeColUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE
    ASSOCIATION_CLASS: ClassVar = model.CaseTypeColSetMember
    LINK_FIELD_NAME1: ClassVar = "case_type_col_set_id"
    LINK_FIELD_NAME2: ClassVar = "case_type_col_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.CaseTypeColSetMember]


class CaseSetCreateCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_set: model.CaseSet = Field(description="The case set to create.")
    data_collection_ids: set[UUID] = Field(
        description="The data collections to associate with the case set, other than the created_in_data_collection. The latter will be removed from the set if present.",
    )
    case_ids: set[UUID] | None = Field(
        description="The cases to associate with the case set upon creation, if any. These cases must have the same case type as the case set.",
        default=None,
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        self.data_collection_ids.discard(self.case_set.created_in_data_collection_id)
        return self


class CasesCreateCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    cases: list[model.Case] = Field(
        description="The cases to create. All cases must have the same case type and created_in_data_collection."
    )
    data_collection_ids: set[UUID] = Field(
        description="The data collections to associate with the cases, other than the created_in_data_collection. The latter will be removed from the set if present."
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if len(set(x.case_type_id for x in self.cases)) > 1:
            raise ValueError("Not all cases have the same case type.")
        case_ids = set()
        for i, case in enumerate(self.cases):
            if case.id in case_ids:
                raise ValueError(f"Duplicate case id: {case.id}")
            if case.id is not None:
                case_ids.add(case.id)
        if self.cases:
            self.data_collection_ids.discard(
                self.cases[0].created_in_data_collection_id
            )
        return self


class RetrieveCaseSetStatsCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_set_ids: list[UUID] | None = Field(
        default=None,
        description="The case set ids to retrieve stats for, if not all. UNIQUE",
    )

    @field_validator("case_set_ids", mode="after")
    def _validate_case_set_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case set ids")
        return value


class RetrieveCaseTypeStatsCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_type_ids: set[UUID] | None = Field(
        default=None,
        description="The case type ids to retrieve stats for, if not all.",
    )
    datetime_range_filter: TypedDatetimeRangeFilter | None = Field(
        default=None,
        description="The datetime range to filter cases by, if any. The key attribute fo the filter should be left empty.",
    )

    @field_validator("case_type_ids", mode="after")
    def _validate_case_type_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case type ids")
        return value


class RetrieveCompleteCaseTypeCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_type_id: UUID


class RetrieveCasesByQueryCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_query: model.CaseQuery


class RetrieveCasesByIdCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_ids: list[UUID] = Field(
        description="The case ids to retrieve cases for. UNIQUE"
    )

    @field_validator("case_ids", mode="after")
    def _validate_case_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case ids")
        return value


class RetrieveCaseRightsCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_ids: list[UUID] = Field(
        description="The case ids to retrieve access for. UNIQUE"
    )

    @field_validator("case_ids", mode="after")
    def _validate_case_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case ids")
        return value


class RetrieveCaseSetRightsCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    case_set_ids: list[UUID] = Field(
        description="The case set ids to retrieve access for. UNIQUE"
    )

    @field_validator("case_set_ids", mode="after")
    def _validate_case_set_ids(cls, value: list[UUID]) -> list[UUID]:
        if len(set(value)) < len(value):
            raise ValueError("Duplicate case set ids")
        return value


class RetrievePhylogeneticTreeBySequencesCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE
    tree_algorithm_code: enum.TreeAlgorithmType
    seqdb_seq_distance_protocol_id: UUID
    sequence_ids: list[UUID]


class RetrievePhylogeneticTreeByCasesCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE
    tree_algorithm: enum.TreeAlgorithmType
    genetic_distance_case_type_col_id: UUID
    case_ids: list[UUID]


class RetrieveGeneticSequenceByCaseCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    genetic_sequence_case_type_col_id: UUID
    case_ids: list[UUID]


class RetrieveAlleleProfileCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.CASE

    sequence_ids: list[UUID]


# seq
class RetrieveGeneticSequenceByIdCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQDB

    seq_ids: list[UUID]


# abac
class RetrieveCompleteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ABAC


class RetrieveOrganizationAdminNameEmailsCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ABAC


class UpdateUserOwnOrganizationCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ABAC

    organization_id: UUID
    is_new_user: bool = False


# ontology
class ConceptSetConceptUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ONTOLOGY
    ASSOCIATION_CLASS: ClassVar = model.ConceptSetMember
    LINK_FIELD_NAME1: ClassVar = "concept_set_id"
    LINK_FIELD_NAME2: ClassVar = "concept_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.ConceptSetMember]


class DiseaseEtiologicalAgentUpdateAssociationCommand(UpdateAssociationCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ONTOLOGY
    ASSOCIATION_CLASS: ClassVar = model.Etiology
    LINK_FIELD_NAME1: ClassVar = "disease_id"
    LINK_FIELD_NAME2: ClassVar = "etiological_agent_id"

    obj_id1: UUID | None = None
    obj_id2: UUID | None = None
    association_objs: list[model.Etiology]


# system
class RetrieveOutagesCommand(Command):
    SERVICE_TYPE = enum.ServiceType.SYSTEM


# rbac
class GetOwnPermissionsCommand(Command):
    SERVICE_TYPE = enum.ServiceType.RBAC


DOMAIN.register_locals(locals())



from __future__ import annotations

import logging
import uuid
from typing import Callable, Hashable
from uuid import UUID

from gen_epix.casedb.domain import command, enum, model
from gen_epix.casedb.domain.service import BaseRbacService
from gen_epix.fastapp import App, Command, Permission


class RbacService(BaseRbacService):

    def __init__(
        self,
        app: App,
        logger: logging.Logger | None = None,
        **kwargs: dict,
    ):
        kwargs["id_factory"] = kwargs.get("id_factory", uuid.uuid4)  # type: ignore[arg-type]
        kwargs["service_type"] = kwargs.get(
            "service_type", BaseRbacService.SERVICE_TYPE  # type: ignore[arg-type]
        )
        super().__init__(app, logger=logger, **kwargs)  # type: ignore[arg-type]
        self._id_factory: Callable[[], UUID]
        # Register permissions without RBAC
        for (
            command_class,
            permission_type,
        ) in BaseRbacService.NO_RBAC_PERMISSIONS:
            permission = self.app.domain.get_permission(command_class, permission_type)
            self.register_permission_without_rbac(permission)

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)
        f(command.GetOwnPermissionsCommand, self.retrieve_user_permissions)

    def register_policies(self) -> None:
        self.register_rbac_policies()

    def retrieve_user_roles(self, user: model.User) -> set[Hashable]:  # type: ignore[override]
        return user.roles  # type: ignore[return-value]

    def retrieve_user_is_non_rbac_authorized(self, cmd: Command) -> bool:
        """
        Check if the user is authorized to perform the command in addition to RBAC.
        A root user is always authorized, any other user must have is_active=True.
        """
        user: model.User | None = cmd.user  # type: ignore[assignment]
        if user is None:
            return False
        return user.is_active or enum.Role.ROOT in user.roles

    def retrieve_user_is_root(self, user: model.User) -> bool:  # type: ignore[override]
        return enum.Role.ROOT in user.roles

    def get_own_permissions(
        self, cmd: command.GetOwnPermissionsCommand
    ) -> set[Permission]:
        user: model.User | None = cmd.user
        if not user or not user.id:
            return set()
        return self.retrieve_user_permissions(user)  # type: ignore[arg-type]



# pylint: disable=useless-import-alias
from typing import Type

from gen_epix.fastapp import BaseService
from gen_epix.fastapp.services.auth import BaseAuthService as BaseAuthService
from gen_epix.seqdb.domain import enum
from gen_epix.seqdb.domain.service.abac import BaseAbacService as BaseAbacService
from gen_epix.seqdb.domain.service.organization import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.seqdb.domain.service.rbac import BaseRbacService as BaseRbacService
from gen_epix.seqdb.domain.service.seq import BaseSeqService as BaseSeqService
from gen_epix.seqdb.domain.service.system import BaseSystemService as BaseSystemService

ORDERED_SERVICE_TYPES: list[enum.ServiceType] = [
    enum.ServiceType.ORGANIZATION,
    enum.ServiceType.AUTH,
    enum.ServiceType.ABAC,
    enum.ServiceType.SYSTEM,
    enum.ServiceType.SEQ,
    enum.ServiceType.RBAC,
]

BASE_SERVICE_CLASS_MAP: dict[enum.ServiceType, Type[BaseService]] = {
    enum.ServiceType.ORGANIZATION: BaseOrganizationService,
    enum.ServiceType.AUTH: BaseAuthService,
    enum.ServiceType.ABAC: BaseAbacService,
    enum.ServiceType.SYSTEM: BaseSystemService,
    enum.ServiceType.SEQ: BaseSeqService,
    enum.ServiceType.RBAC: BaseRbacService,
}



from enum import Enum
from typing import Any
from uuid import UUID

from pydantic import BaseModel, Field

from gen_epix.casedb.domain import DOMAIN, enum, model
from gen_epix.fastapp import Permission as ServicePermission
from gen_epix.fastapp import PermissionType

CommandName = Enum("CommandName", {x: x for x in DOMAIN.command_names})  # type: ignore[misc]


class Model(BaseModel):
    id: UUID | None = Field(
        default=None,
        description="The unique identifier for the obj.",
    )


# TODO: investigate why just inheriting from the domain model (here ServicePermission) no longer works, as it raises a a Pydantic exception around ClassVar
# Workaround is to create the entire API model from scratch with exactly the same fields as the domain model and the relevant fields changed (here command_name)
# This means that the API model has to be kept in sync with the domain model, which is not ideal
# class Permission(ServicePermission):
class Permission(BaseModel, frozen=True):
    command_name: CommandName  # type: ignore
    permission_type: PermissionType

    @staticmethod
    def _map_model(obj: Any, map_to: bool) -> dict:
        fun = (lambda x: x.value) if map_to else (lambda x: CommandName[x])
        return {
            x: (getattr(obj, x) if x != "command_name" else fun(obj.command_name))
            for x in obj.model_fields.keys()
        }

    @staticmethod
    def from_model(permission: ServicePermission | None) -> Any:
        return (
            None
            if permission is None
            else Permission(**Permission._map_model(permission, False))
        )

    @staticmethod
    def to_model(permission: Any) -> ServicePermission | None:
        return (
            None
            if permission is None
            else ServicePermission(**Permission._map_model(permission, True))
        )


# class CompleteUser(model.CompleteUser):
class CompleteUser(Model):
    email: str = Field(description="The email of the user, UNIQUE", max_length=320)
    name: str | None = Field(
        default=None, description="The full name of the user", max_length=255
    )
    is_active: bool = Field(
        default=True,
        description="Whether the user is active or not. An inactive user cannot perform any actions that require authorization.",
    )
    organization_id: UUID = Field(
        description="The ID of the organization of the user. FOREIGN KEY"
    )
    organization: model.Organization | None
    permissions: set[Permission]
    roles: set[enum.Role]
    # case_abac: model.CaseAbac = Field(
    #     description="The case abac of the user. This is used to determine the user's access to cases.",
    # )

    @staticmethod
    def _map_model(obj: Any, map_to: bool) -> dict:
        fun = Permission.to_model if map_to else Permission.from_model
        return {
            x: (
                {fun(y) for y in obj.permissions}
                if x == "permissions"
                else getattr(obj, x)
            )
            for x in obj.model_fields.keys()
        }

    @staticmethod
    def from_model(
        complete_user: model.CompleteUser | None,
    ) -> Any:
        return (
            None
            if complete_user is None
            else CompleteUser(**CompleteUser._map_model(complete_user, False))
        )

    @staticmethod
    def to_model(
        complete_user: Any,
    ) -> model.CompleteUser | None:
        return (
            None
            if complete_user is None
            else model.CompleteUser(**CompleteUser._map_model(complete_user, True))
        )


# Set the API model classes for the domain models
model.CompleteUser.ENTITY.set_create_api_model_class(CompleteUser)
model.CompleteUser.ENTITY.set_read_api_model_class(CompleteUser)



from typing import Any

from gen_epix.fastapp import App, exc
from gen_epix.fastapp.enum import AuthProtocol
from gen_epix.fastapp.services.auth.idp_client import IDPClient
from gen_epix.fastapp.services.auth.literal import EMAIL_PATTERN
from gen_epix.fastapp.services.auth.model import OIDCConfiguration
from gen_epix.fastapp.services.auth.oidc_client import OIDCClient

OIDC_DISCOVERY_DOCUMENT_KEYS = {
    "issuer",
    "authorization_endpoint",
    "token_endpoint",
    "jwks_uri",
    "userinfo_endpoint",
    "response_types_supported",
    "subject_types_supported",
    "id_token_signing_alg_values_supported",
}


def get_email_from_claims(
    claims: dict[str, Any],
) -> str | None:
    if "email" in claims and isinstance(claims["email"], str):
        return claims["email"].lower()
    for claim in claims.values():
        if isinstance(claim, str) and EMAIL_PATTERN.match(claim.lower()):
            return claim.lower()
    return None


def create_idp_clients_from_config(
    app: App, idps_cfg: list[dict[str, str | list]] | None
) -> list[IDPClient]:
    if not idps_cfg:
        idps_cfg = []
    idp_clients: list[IDPClient] = []
    idp_names = set()
    idp_labels = set()
    logger = app.logger
    for idp_cfg in idps_cfg:
        idp_name = idp_cfg["name"]
        idp_label = idp_cfg["label"]
        if idp_name in idp_names or idp_label in idp_labels:
            msg = (
                "Authentication service name and/or label are not unique: "
                f"{idp_cfg['name']}, {idp_cfg['label']}"
            )
            if logger:
                logger.error(app.create_log_message("30a9a272", msg))
            raise exc.InitializationServiceError(msg)
        idp_names.add(idp_name)
        idp_labels.add(idp_label)

        try:
            protocol = AuthProtocol[str(idp_cfg["protocol"])]
            if protocol == AuthProtocol.OIDC:
                discovery_document = (
                    idp_cfg
                    if OIDC_DISCOVERY_DOCUMENT_KEYS.issubset(set(idp_cfg.keys()))
                    else None
                )
                idp_client = OIDCClient(
                    OIDCConfiguration(**idp_cfg),  # type: ignore
                    logger=logger,
                    log_item_class=app.log_item_class,
                    discovery_document=discovery_document,
                )
            else:
                raise exc.InitializationServiceError(
                    f"Protocol {protocol.value} not implemented"
                )
            idp_clients.append(idp_client)
        except Exception as exception:
            # Unable to initialize authentication service: do not raise
            # an error to avoid entire app not starting up
            msg = "Could not initialize authentication service " f"{idp_cfg['name']}"
            if logger:
                logger.error(
                    app.create_log_message("48b7e021", msg, exception=exception)
                )
    for idp_client in idp_clients:  # type: ignore
        if isinstance(idp_client, OIDCClient):
            if logger:
                logger.info(
                    app.create_log_message(
                        "7e0b64cc",
                        f"OIDC service on {idp_client._issuer} initialized",
                    )
                )
        else:
            raise exc.InitializationServiceError(
                f"Authentication service of type {type(idp_client)} " "not implemented"
            )
    return idp_clients



from gen_epix.fastapp import PermissionTypeSet
from gen_epix.fastapp.services.rbac import BaseRbacService
from gen_epix.omopdb.domain import command
from gen_epix.omopdb.domain.enum import Role


class RoleGenerator:

    ROLE_PERMISSIONS = {
        Role.ADMIN: {  # type: ignore
            # organization
            (command.IdentifierIssuerCrudCommand, PermissionTypeSet.CU),
            (command.UserCrudCommand, PermissionTypeSet.R),
            (command.UserInvitationCrudCommand, PermissionTypeSet.CRD),
            (command.InviteUserCommand, PermissionTypeSet.E),
            (command.UpdateUserCommand, PermissionTypeSet.E),
            (command.OutageCrudCommand, PermissionTypeSet.CRUD),
            (command.DataCollectionCrudCommand, PermissionTypeSet.CRU),
            (command.DataCollectionSetCrudCommand, PermissionTypeSet.CRUD),
            (command.DataCollectionSetMemberCrudCommand, PermissionTypeSet.CRUD),
        },
        Role.REFDATA_ADMIN: {},  # type: ignore
        Role.DATA_READER: {},  # type: ignore
        Role.DATA_WRITER: {},  # type: ignore
        Role.GUEST: set(),  # type: ignore
    }

    # Tree hierarchy of roles: each role can do everything the roles below it can do.
    # Hierarchy described here per role with union of all roles below it.
    ROLE_HIERARCHY = {
        Role.ROOT: {  # type: ignore
            Role.ADMIN,
            Role.REFDATA_ADMIN,
            Role.DATA_READER,
            Role.DATA_WRITER,
            Role.GUEST,
        },
        Role.ADMIN: {  # type: ignore
            Role.REFDATA_ADMIN,
            Role.DATA_READER,
            Role.DATA_WRITER,
            Role.GUEST,
        },
        Role.REFDATA_ADMIN: {Role.GUEST},  # type: ignore
        Role.DATA_READER: {Role.GUEST},  # type: ignore
        Role.DATA_WRITER: {Role.DATA_READER, Role.GUEST},  # type: ignore
        Role.GUEST: set(),  # type: ignore
    }

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSIONS
    )



from typing import Callable
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator
from gen_epix.seqdb.api.base import EXCLUDED_PERMISSIONS
from gen_epix.seqdb.domain import command, enum, model


class RetrievePhylogeneticTreeRequestBody(PydanticBaseModel):
    seq_distance_protocol_id: UUID
    tree_algorithm: enum.TreeAlgorithm
    seq_ids: list[UUID]
    leaf_codes: list[str] | None


class RetrieveSeqRequestBody(PydanticBaseModel):
    seq_ids: list[UUID]


class RetrieveAlleleProfileRequestBody(PydanticBaseModel):
    seq_ids: list[UUID]
    locus_set_id: UUID


def create_seq_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    **kwargs: dict,
) -> None:
    assert handle_exception

    @router.post(
        "/retrieve/phylogenetic_tree", operation_id="retrieve__phylogenetic_tree"
    )
    async def retrieve__phylogenetic_tree(
        user: registered_user_dependency, request_body: RetrievePhylogeneticTreeRequestBody  # type: ignore
    ) -> model.PhylogeneticTree:
        try:
            retval: model.PhylogeneticTree = app.handle(
                command.RetrievePhylogeneticTreeCommand(
                    user=user,
                    seq_distance_protocol_id=request_body.seq_distance_protocol_id,
                    tree_algorithm=request_body.tree_algorithm,
                    seq_ids=request_body.seq_ids,
                    leaf_names=request_body.leaf_codes,
                )
            )
        except Exception as exception:
            handle_exception("dc71bce0", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    @router.post("/retrieve/seq", operation_id="retrieve__seq")
    async def retrieve__seq(
        user: registered_user_dependency, request_body: RetrieveSeqRequestBody  # type: ignore
    ) -> list[model.CompleteSeq]:
        try:
            retval: list[model.CompleteSeq] = app.handle(
                command.RetrieveCompleteSeqCommand(
                    user=user,
                    seq_ids=request_body.seq_ids,
                )
            )
        except Exception as exception:
            handle_exception("ac218f73", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    @router.post("/retrieve/allele_profile", operation_id="retrieve__allele_profile")
    async def retrieve__allele_profile(
        user: registered_user_dependency, request_body: RetrieveAlleleProfileRequestBody  # type: ignore
    ) -> list[model.CompleteAlleleProfile]:
        try:
            retval: list[model.CompleteAlleleProfile] = app.handle(
                command.RetrieveCompleteAlleleProfileCommand(
                    user=user,
                    seq_ids=request_body.seq_ids,
                    locus_set_id=request_body.locus_set_id,
                )
            )
        except Exception as exception:
            handle_exception("f1d282b4", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.SEQ,
        user_dependency=registered_user_dependency,
        excluded_permissions=EXCLUDED_PERMISSIONS,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



import datetime
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy_utils.types.uuid import UUIDType

from gen_epix.fastapp.repositories.sa import ServerUtcCurrentTime
from gen_epix.seqdb.domain import DOMAIN


class RowMetadataMixin:
    id: Mapped[UUID] = mapped_column(UUIDType(), primary_key=True)
    _created_at: Mapped[datetime.datetime] = mapped_column(
        sa.DateTime, nullable=False, server_default=ServerUtcCurrentTime()
    )
    _modified_at: Mapped[datetime.datetime] = mapped_column(
        sa.DateTime,
        nullable=False,
        server_default=ServerUtcCurrentTime(),
        onupdate=ServerUtcCurrentTime(),
    )
    _modified_by: Mapped[UUID] = mapped_column(UUIDType(), nullable=True)
    _version: Mapped[int] = mapped_column(sa.Integer, nullable=False, default=1)
    __mapper_args__ = {"version_id_col": _version}


class CodeMixin:
    code: Mapped[str] = mapped_column(sa.String, nullable=False)


class QualityMixin:
    quality_score: Mapped[float] = mapped_column(sa.Float, nullable=True)
    quality: Mapped[str] = mapped_column(sa.String, nullable=True)


class SeqMixin:
    seq: Mapped[str] = mapped_column(sa.String, nullable=False)
    seq_format: Mapped[str] = mapped_column(sa.String, nullable=False)
    seq_hash_sha256: Mapped[bytes] = mapped_column(
        sa.LargeBinary(length=32), nullable=False
    )
    length: Mapped[int] = mapped_column(sa.Integer, nullable=False)


class AlignmentMixin:
    aln: Mapped[str] = mapped_column(sa.String, nullable=False)
    aln_format: Mapped[str] = mapped_column(sa.String, nullable=False)
    aln_hash_sha256: Mapped[bytes] = mapped_column(
        sa.LargeBinary(length=32), nullable=False
    )


class ProtocolMixin:
    code: Mapped[str] = mapped_column(sa.String, nullable=False)
    name: Mapped[str] = mapped_column(sa.String, nullable=False)
    version: Mapped[str] = mapped_column(sa.String, nullable=True)
    description: Mapped[str] = mapped_column(sa.String, nullable=True)
    props: Mapped[dict[str, str]] = mapped_column(sa.JSON(), nullable=True)


SERVICE_METADATA_FIELDS = {
    model_class: ["_modified_by"] for model_class in DOMAIN.models
}

DB_METADATA_FIELDS = {
    model_class: [
        "_created_at",
        "_modified_at",
        "_version",
    ]
    for model_class in DOMAIN.models
}

GENERATE_SERVICE_METADATA = {
    model_class: lambda x, y: {
        "_modified_by": y,
    }
    for model_class in DOMAIN.models
}



from typing import Callable, Type

from gen_epix.fastapp.domain.key import Key
from gen_epix.fastapp.domain.link import Link


def create_keys(keys: dict[int, Key | str | tuple | Callable]) -> dict[int, Key]:
    retval = {}
    for x, y in keys.items():
        if isinstance(y, Key):
            retval[x] = y
        else:
            retval[x] = Key(y)
    return retval


def create_links(
    links: dict[int, Link | tuple[str, Type, str | None]],
) -> dict[int, Link]:
    retval = {}
    for x, y in links.items():
        if isinstance(y, Link):
            retval[x] = y
        else:
            retval[x] = Link(
                link_field_name=y[0],
                link_model_class=y[1],
                relationship_field_name=y[2],
            )
    return retval



# pylint: disable=useless-import-alias
from gen_epix.casedb.services.abac import AbacService as AbacService
from gen_epix.casedb.services.case import CaseService as CaseService
from gen_epix.casedb.services.geo import GeoService as GeoService
from gen_epix.casedb.services.ontology import OntologyService as OntologyService
from gen_epix.casedb.services.organization import (
    OrganizationService as OrganizationService,
)
from gen_epix.casedb.services.rbac import RbacService as RbacService
from gen_epix.casedb.services.seqdb import SeqdbService as SeqdbService
from gen_epix.casedb.services.subject import SubjectService as SubjectService
from gen_epix.casedb.services.system import SystemService as SystemService
from gen_epix.casedb.services.user_manager import UserManager as UserManager



from typing import Type

from gen_epix.casedb.domain import command
from gen_epix.casedb.domain.enum import Role
from gen_epix.fastapp import PermissionTypeSet
from gen_epix.fastapp.services.rbac import BaseRbacService


class RoleGenerator:

    ROLE_PERMISSION_SETS: dict[
        Role, set[tuple[Type[command.Command], PermissionTypeSet]]
    ] = {
        # TODO: remove UPDATE from association objects that do not have properties of their own such as CaseTypeSetMember
        Role.APP_ADMIN: {
            # case
            (
                command.CaseSetCrudCommand,
                PermissionTypeSet.C,
            ),  # Other users can only use dedicated command
            (
                command.CaseCrudCommand,
                PermissionTypeSet.C,
            ),  # Other users can only use dedicated command
            (command.CaseTypeCrudCommand, PermissionTypeSet.D),
            (command.CaseSetCategoryCrudCommand, PermissionTypeSet.CU),
            (command.CaseSetStatusCrudCommand, PermissionTypeSet.CU),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.D),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.D),
            (
                command.DataCollectionSetDataCollectionUpdateAssociationCommand,  # type: ignore[arg-type]
                PermissionTypeSet.E,
            ),
            # abac
            (command.OrganizationAdminPolicyCrudCommand, PermissionTypeSet.CUD),
            (command.OrganizationAccessCasePolicyCrudCommand, PermissionTypeSet.CUD),
            (
                command.OrganizationShareCasePolicyCrudCommand,
                PermissionTypeSet.CUD,
            ),
            # organization
            (command.OrganizationCrudCommand, PermissionTypeSet.CU),
            (
                command.OrganizationSetOrganizationUpdateAssociationCommand,  # type: ignore[arg-type]
                PermissionTypeSet.E,
            ),
            (command.DataCollectionCrudCommand, PermissionTypeSet.CU),
            (
                command.DataCollectionSetCrudCommand,
                PermissionTypeSet.CRUD,
            ),  # TODO: READ permission can be set broader once this entity is actually used
            (
                command.DataCollectionSetMemberCrudCommand,
                PermissionTypeSet.CRUD,
            ),  # TODO: READ permission can be set broader once this entity is actually used
            # system
            (command.OutageCrudCommand, PermissionTypeSet.CRUD),
        },
        Role.METADATA_ADMIN: {
            # case
            (command.GeneticDistanceProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColSetCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeColSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.CaseTypeCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetCaseTypeUpdateAssociationCommand, PermissionTypeSet.E),  # type: ignore[arg-type]
            (
                command.CaseTypeColSetCaseTypeColUpdateAssociationCommand,  # type: ignore[arg-type]
                PermissionTypeSet.E,
            ),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.CRU),
            (command.CaseTypeSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.ColCrudCommand, PermissionTypeSet.CRU),
            (command.DimCrudCommand, PermissionTypeSet.CRU),
            # ontology
            (command.ConceptCrudCommand, PermissionTypeSet.CRU),
            (command.ConceptSetConceptUpdateAssociationCommand, PermissionTypeSet.E),  # type: ignore[arg-type]
            (command.ConceptSetCrudCommand, PermissionTypeSet.CRU),
            (command.ConceptSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.DiseaseCrudCommand, PermissionTypeSet.CRU),
            (
                command.DiseaseEtiologicalAgentUpdateAssociationCommand,  # type: ignore[arg-type]
                PermissionTypeSet.E,
            ),
            (command.EtiologicalAgentCrudCommand, PermissionTypeSet.CRU),
            (command.EtiologyCrudCommand, PermissionTypeSet.CRU),
            (command.RegionCrudCommand, PermissionTypeSet.CRU),
            (command.RegionSetCrudCommand, PermissionTypeSet.CRU),
            (command.RegionSetShapeCrudCommand, PermissionTypeSet.CRUD),
            # organization
            (command.UserCrudCommand, PermissionTypeSet.R),
            (command.DataCollectionCrudCommand, PermissionTypeSet.R),
        },
        Role.ORG_ADMIN: {
            # organization
            (command.InviteUserCommand, PermissionTypeSet.E),
            (command.UpdateUserCommand, PermissionTypeSet.E),
            (command.UserInvitationCrudCommand, PermissionTypeSet.CRD),
            # abac
            (command.UserAccessCasePolicyCrudCommand, PermissionTypeSet.CUD),
            (command.UserShareCasePolicyCrudCommand, PermissionTypeSet.CUD),
        },
        Role.ORG_USER: {
            # case
            (command.GeneticDistanceProtocolCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.R),
            (command.CaseSetCategoryCrudCommand, PermissionTypeSet.R),
            (command.CaseSetDataCollectionLinkCrudCommand, PermissionTypeSet.CRUD),
            (command.CaseSetStatusCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeColSetCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetCategoryCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetCrudCommand, PermissionTypeSet.R),
            (command.CaseTypeSetMemberCrudCommand, PermissionTypeSet.R),
            (
                command.CaseDataCollectionLinkCrudCommand,
                PermissionTypeSet.CRUD,
            ),
            (command.CaseSetCreateCommand, PermissionTypeSet.E),
            (command.CasesCreateCommand, PermissionTypeSet.E),
            (command.CaseSetCrudCommand, PermissionTypeSet.RUD),
            (command.CaseSetMemberCrudCommand, PermissionTypeSet.CRUD),
            (command.RetrieveCaseSetStatsCommand, PermissionTypeSet.E),
            (command.RetrieveCaseTypeStatsCommand, PermissionTypeSet.E),
            (command.RetrieveCasesByIdCommand, PermissionTypeSet.E),
            (command.RetrieveCasesByQueryCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteCaseTypeCommand, PermissionTypeSet.E),
            (command.RetrieveCaseSetRightsCommand, PermissionTypeSet.E),
            (command.RetrieveCaseRightsCommand, PermissionTypeSet.E),
            # ontology
            (command.ConceptCrudCommand, PermissionTypeSet.R),
            (command.ConceptSetCrudCommand, PermissionTypeSet.R),
            (command.ConceptSetMemberCrudCommand, PermissionTypeSet.R),
            (command.DiseaseCrudCommand, PermissionTypeSet.R),
            (command.EtiologicalAgentCrudCommand, PermissionTypeSet.R),
            (command.EtiologyCrudCommand, PermissionTypeSet.R),
            (command.RegionSetCrudCommand, PermissionTypeSet.R),
            (command.RegionSetShapeCrudCommand, PermissionTypeSet.R),
            (command.RegionCrudCommand, PermissionTypeSet.R),
            # organization
            (command.UserCrudCommand, PermissionTypeSet.R),
            (command.DataCollectionCrudCommand, PermissionTypeSet.R),
            (command.DataCollectionRelationCrudCommand, PermissionTypeSet.R),
            (command.OrganizationCrudCommand, PermissionTypeSet.R),
            (command.RetrieveOrganizationAdminNameEmailsCommand, PermissionTypeSet.E),
            (command.RetrieveOrganizationContactCommand, PermissionTypeSet.E),
            (command.UpdateUserOwnOrganizationCommand, PermissionTypeSet.E),
            # abac
            (command.OrganizationAdminPolicyCrudCommand, PermissionTypeSet.R),
            (command.OrganizationAccessCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.OrganizationShareCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.UserAccessCasePolicyCrudCommand, PermissionTypeSet.R),
            (command.UserShareCasePolicyCrudCommand, PermissionTypeSet.R),
            # seq
            (command.RetrieveAlleleProfileCommand, PermissionTypeSet.E),
            (command.RetrieveGeneticSequenceByCaseCommand, PermissionTypeSet.E),
            (command.RetrievePhylogeneticTreeByCasesCommand, PermissionTypeSet.E),
            (command.RetrievePhylogeneticTreeBySequencesCommand, PermissionTypeSet.E),
        },
        Role.GUEST: {
            # organization
            (command.RetrieveCompleteUserCommand, PermissionTypeSet.E),
            # system
            (command.RetrieveOutagesCommand, PermissionTypeSet.E),
        },
    }

    # Tree hierarchy of roles: each role can do everything the roles below it can do.
    # Hierarchy described here per role with union of all roles below it.
    ROLE_HIERARCHY: dict[Role, set[Role]] = {
        Role.ROOT: {
            Role.APP_ADMIN,
            Role.ORG_ADMIN,
            Role.METADATA_ADMIN,
            Role.ORG_USER,
            Role.GUEST,
        },
        Role.APP_ADMIN: {
            Role.ORG_ADMIN,
            Role.METADATA_ADMIN,
            Role.ORG_USER,
            Role.GUEST,
        },
        Role.ORG_ADMIN: {
            Role.ORG_USER,
            Role.GUEST,
        },
        Role.METADATA_ADMIN: {Role.GUEST},
        Role.ORG_USER: {Role.GUEST},
        Role.GUEST: set(),
    }

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSION_SETS  # type: ignore[arg-type]
    )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar

from gen_epix.omopdb.domain import DOMAIN, model
from gen_epix.omopdb.domain.command.base import CrudCommand


# organization
class IdentifierIssuerCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.IdentifierIssuer


class DataCollectionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollection


class DataCollectionSetCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSet


class DataCollectionSetMemberCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DataCollectionSetMember


class UserCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.User


class UserInvitationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserInvitation


# system
class OutageCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Outage


# omop

# The content of this file is to be inserted into the gen_epix/omopdb/domain/command/crud.py file
# under the "omop" section. This file was generated by the generate_omop_pydantic_model_defs.py script.

# omop


class PersonCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Person


class ObservationPeriodCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ObservationPeriod


class VisitOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.VisitOccurrence


class VisitDetailCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.VisitDetail


class ConditionOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConditionOccurrence


class DrugExposureCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugExposure


class ProcedureOccurrenceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ProcedureOccurrence


class DeviceExposureCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DeviceExposure


class MeasurementCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Measurement


class ObservationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Observation


class NoteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Note


class NoteNlpCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.NoteNlp


class SpecimenCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Specimen


class FactRelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.FactRelationship


class SurveyConductCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SurveyConduct


class LocationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Location


class LocationHistoryCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.LocationHistory


class CareSiteCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CareSite


class ProviderCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Provider


class PayerPlanPeriodCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.PayerPlanPeriod


class CostCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Cost


class DrugEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugEra


class DoseEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DoseEra


class ConditionEraCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConditionEra


class MetadataCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Metadata


class CdmSourceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CdmSource


class ConceptCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Concept


class VocabularyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Vocabulary


class DomainCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Domain


class ConceptClassCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptClass


class ConceptRelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptRelationship


class RelationshipCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Relationship


class ConceptSynonymCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptSynonym


class ConceptAncestorCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.ConceptAncestor


class SourceToConceptMapCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.SourceToConceptMap


class DrugStrengthCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.DrugStrength


class CohortCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Cohort


class CohortDefinitionCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.CohortDefinition


class ProvenanceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Provenance


class EtlCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Etl


class SourceCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Source


class OrganizationCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.Organization


DOMAIN.register_locals(locals())



from gen_epix.omopdb.domain import DOMAIN, enum
from gen_epix.omopdb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.omopdb.repositories.sa_model.organization import *
from util.util import set_entity_repository_model_classes

FIELD_NAME_MAP: dict[Type, dict[str, str]] = {}

set_entity_repository_model_classes(
    DOMAIN,
    enum.ServiceType,
    RowMetadataMixin,
    "gen_epix.omopdb.repositories.sa_model",
    field_name_map=FIELD_NAME_MAP,
)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar
from uuid import UUID

from pydantic import Field, field_serializer

from gen_epix.casedb.domain import DOMAIN, enum
from gen_epix.casedb.domain.model.base import Model
from gen_epix.casedb.domain.model.case.case import (
    GeneticDistanceProtocol,
    TreeAlgorithm,
)
from gen_epix.fastapp import Entity

_SERVICE_TYPE = enum.ServiceType.SEQDB
_ENTITY_KWARGS = {
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class GeneticSequence(Model):
    """
    A class representing a genetic sequence. Temporary implementation.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="genetic_sequences",
        table_name="genetic_sequence",  # TODO: temporarily added for extraction of seqdb demo data
        persistable=False,
        **_ENTITY_KWARGS,
    )
    nucleotide_sequence: str | None = Field(
        default=None, description="The nucleotide sequence"
    )
    distances: dict[UUID, float] | None = Field(
        default=None, description="The distances to other sequences"
    )

    @field_serializer("distances")
    def serialize_distances(self, value: dict[UUID, float], _info):
        return {str(x): y for x, y in value.items()}


class AlleleProfile(Model):
    """
    A class representing an allele profile. Temporary implementation.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="allele_profiles",
        **_ENTITY_KWARGS,
    )
    # TODO: add link to sequence and gene set
    allele_profile: str | None = Field(default=None, description="The allele profile")


class PhylogeneticTree(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="phylogenetic_trees",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    tree_algorithm_id: UUID | None = Field(
        default=None, description="The ID of the tree algorithm. FOREIGN KEY"
    )
    tree_algorithm: TreeAlgorithm = Field(
        default=None, description="The tree algorithm"
    )
    tree_algorithm_code: enum.TreeAlgorithmType = Field(
        description="The tree algorithm"
    )
    genetic_distance_protocol_id: UUID | None = Field(
        default=None, description="The ID of the genetic distance protocol. FOREIGN KEY"
    )
    genetic_distance_protocol: GeneticDistanceProtocol = Field(
        default=None, description="The genetic distance protocol"
    )
    leaf_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the leaves of the phylogenetic tree.",
    )
    sequence_ids: list[UUID] | None = Field(
        default=None,
        description="The list of unique identifiers of the sequence of each leaf of the phylogenetic tree.",
    )
    newick_repr: str = Field(
        description="The Newick representation of the phylogenetic tree."
    )


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



from datetime import date, datetime

from pydantic import BaseModel

from gen_epix.omopdb.services.model_anonymizer import ModelAnonymizer


class Model(BaseModel):
    str_value: str | None
    int_value: int | None
    float_value: float | None
    bool_value: bool | None
    date_value: date | None
    datetime_value: datetime | None
    list_value: list | None
    dict_value: dict | None
    tuple_value: tuple | None
    set_value: set | None
    bytes_value: bytes | None


class TestAnonymizer:
    def test_anonymize(self) -> None:
        model_anonymizer = ModelAnonymizer()
        model = Model(
            str_value="test",
            int_value=123,
            float_value=123.456,
            bool_value=True,
            date_value=date(2023, 1, 1),
            datetime_value=datetime(2023, 1, 1, 12, 0),
            list_value=["a", "b", "c"],
            dict_value={"key": "value"},
            tuple_value=("a", "b"),
            set_value={"a", "b"},
            bytes_value=b"test",
        )
        raise NotImplementedError("Test not implemented yet")



import datetime
from typing import Any
from uuid import UUID

from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.service.organization import BaseOrganizationService
from gen_epix.casedb.domain.service.rbac import BaseRbacService
from gen_epix.fastapp import BaseUnitOfWork, BaseUserManager, CrudOperation, Permission
from gen_epix.fastapp.services.auth import get_email_from_claims


class UserManager(BaseUserManager):
    def __init__(
        self,
        organization_service: BaseOrganizationService,
        rbac_service: BaseRbacService,
        root_cfg: dict[str, dict[str, str]],
        automatic_new_user_cfg: dict[str, dict[str, str]] | None = None,
    ):
        self._organization_service: BaseOrganizationService = organization_service
        self._rbac_service: BaseRbacService = rbac_service

        # Generate root model objs
        self._root: dict = {}
        self._root["organization"] = model.Organization(**root_cfg["organization"])  # type: ignore[arg-type]
        self._root["user"] = model.User(
            is_active=True,
            organization_id=self._root["organization"].id,
            **root_cfg["user"],  # type: ignore[arg-type]
        )
        self._root["user"].roles.add(enum.Role.ROOT)

        # Get automatic new user data
        self._automatic_new_user: dict | None = None
        if automatic_new_user_cfg:
            self._automatic_new_user = {}
            self._automatic_new_user["roles"] = {
                enum.Role[x] for x in automatic_new_user_cfg["roles"]
            }
            self._automatic_new_user["organization"] = {
                **automatic_new_user_cfg["organization"]
            }
            self._automatic_new_user["organization"]["id"] = UUID(
                self._automatic_new_user["organization"]["id"]
            )

    def generate_id(self) -> UUID:
        return self._organization_service.generate_id()  # type: ignore[return-value]

    def get_user_key_from_claims(self, claims: dict[str, Any]) -> str | None:
        return get_email_from_claims(claims)

    def get_user_instance_from_claims(  # type: ignore[override]
        self, claims: dict[str, Any]
    ) -> model.User | None:
        if self._automatic_new_user is None:
            return None
        roles = (
            self._automatic_new_user["roles"]
            if self._automatic_new_user
            else {enum.Role.GUEST}
        )
        organization_id = (
            self._automatic_new_user["organization"]["id"]
            if self._automatic_new_user
            else self._root["organization"].id
        )
        email = get_email_from_claims(claims)
        if not email:
            raise exc.CredentialsAuthError("Email not found in claims")
        return model.User(
            email=email,
            is_active=True,
            roles=roles,
            organization_id=organization_id,
        )

    def is_root_user(self, claims: dict[str, Any]) -> bool:
        user: model.User = self._root["user"]
        return user.email == get_email_from_claims(claims)

    def create_root_user_from_claims(self, claims: dict[str, Any]) -> model.User:  # type: ignore
        assert self._organization_service.repository
        with self._organization_service.repository.uow() as uow:
            # Create root organization if necessary

            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                self._root["organization"].id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                _ = self._organization_service.repository.crud(
                    uow,
                    None,
                    model.Organization,
                    self._root["organization"],
                    None,
                    CrudOperation.CREATE_ONE,
                )

            # Create root user if necessary
            is_existing_root_user = self._organization_service.repository.crud(
                uow,
                None,
                model.User,
                None,
                self._root["user"].id,
                CrudOperation.EXISTS_ONE,
            )
            user: model.User
            if is_existing_root_user:
                user = self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    None,
                    model.User,
                    None,
                    self._root["user"].id,
                    CrudOperation.READ_ONE,
                )
                is_updated = False
                for key, value in claims.items():
                    if not hasattr(user, key) or getattr(user, key) == value:
                        continue
                    is_updated = True
                    if key == "organization_id":
                        user.organization_id = self._root["organization"].id
                    elif key == "roles":
                        user.roles.update(value)
                    else:
                        setattr(user, key, value)
                if enum.Role.ROOT not in user.roles:
                    is_updated = True
                    user.roles.add(enum.Role.ROOT)
                if is_updated:
                    role = self._organization_service.repository.crud(
                        uow,
                        self._root["user"].id,
                        model.User,
                        user,
                        None,
                        CrudOperation.UPDATE_ONE,
                    )
            else:
                user = self._organization_service.repository.crud(  # type:ignore[assignment]
                    uow,
                    self._root["user"].id,
                    model.User,
                    self._root["user"],
                    None,
                    CrudOperation.CREATE_ONE,
                )

        return user

    def create_user_from_claims(self, claims: dict[str, Any]) -> model.User | None:  # type: ignore[override]
        if self._automatic_new_user is None:
            return None
        assert self._organization_service.repository
        organization_id = self._automatic_new_user["organization"]["id"]
        with self._organization_service.repository.uow() as uow:
            # Verify if organization exists
            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                raise exc.InitializationServiceError(
                    "Automatic new user organization does not exist"
                )

            # Verify if user exists and add if not
            # TODO: refactor this to add a separate method for a potential existing user
            is_existing_user = (
                self._organization_service.repository.is_existing_user_by_key(
                    uow, get_email_from_claims(claims)
                )
            )
            if not is_existing_user:
                claims_user = self.get_user_instance_from_claims(claims)
                if not claims_user:
                    raise exc.InitializationServiceError(
                        "Unable to create user from claims"
                    )
                claims_user.id = self.generate_id()
                user: model.User = (
                    self._organization_service.repository.crud(  # type:ignore[assignment]
                        uow,
                        claims_user.id,
                        model.User,
                        claims_user,
                        None,
                        CrudOperation.CREATE_ONE,
                    )
                )

            # Add user case policies by calling switching organization method
            try:
                user = self._organization_service.app.handle(
                    command.UpdateUserOwnOrganizationCommand(
                        user=user,
                        organization_id=user.organization_id,
                        is_new_user=True,
                    ),
                )
            except Exception as exception:
                raise exc.UnauthorizedAuthError("Unable to add user case policies")

        return user

    def create_new_user_from_token(  # type: ignore[override]
        self, new_user: model.User, token: str, **kwargs: dict
    ) -> model.User:
        assert self._organization_service.repository
        created_by_user_id: UUID = kwargs["created_by_user_id"]  # type: ignore[assignment]
        with self._organization_service.repository.uow() as uow:
            # Verify if create_by_user exists and is active
            is_existing_user = self._organization_service.repository.crud(
                uow,
                None,
                model.User,
                None,
                created_by_user_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_user:
                raise exc.UnauthorizedAuthError("Created by user does not exist")
            created_by_user = self.retrieve_user_by_id(created_by_user_id)
            if not created_by_user.is_active:
                raise exc.UnauthorizedAuthError("Created by user is not active")

            # Verify if create_by_user made an invitation for this user that is valid
            timestamp = datetime.datetime.now()
            user_invitations: list[model.UserInvitation] = (
                self._organization_service.repository.crud(  # type: ignore[assignment]
                    uow,
                    created_by_user_id,
                    model.UserInvitation,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                )
            )

            # At least one invitation exists matching the criteria
            user_invitations = [
                x
                for x in user_invitations
                if x.invited_by_user_id == created_by_user_id
                and x.token == token
                and x.email == new_user.email
                and x.organization_id == new_user.organization_id
                and x.expires_at > timestamp
            ]
            if not user_invitations:
                raise exc.UnauthorizedAuthError("Invitation does not exist")

            # Verify if organization exists
            is_existing_organization = self._organization_service.repository.crud(
                uow,
                None,
                model.Organization,
                None,
                new_user.organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                raise exc.UnauthorizedAuthError("Organization does not exist")

            is_existing_user = self.is_existing_user_by_key(new_user.email, uow)
            if is_existing_user:
                raise exc.UnauthorizedAuthError("User already exists")

            try:
                user: model.User = (
                    self._organization_service.repository.crud(  # type:ignore[assignment]
                        uow,
                        created_by_user_id,
                        model.User,
                        model.User(
                            **(new_user.model_dump() | {"id": self.generate_id()})
                        ),
                        None,
                        CrudOperation.CREATE_ONE,
                    )
                )
            except:
                raise exc.UnauthorizedAuthError("Unable to create user")

            return user

    def is_existing_user_by_key(
        self, user_key: str | None, uow: BaseUnitOfWork
    ) -> bool:
        return self._organization_service.repository.is_existing_user_by_key(
            uow, user_key
        )

    def retrieve_user_by_key(self, user_key: str) -> model.User:  # type: ignore[override]
        return self._organization_service.retrieve_user_by_key(user_key)

    def retrieve_user_by_id(self, user_id: UUID) -> model.User:  # type: ignore[override]
        user: model.User = self._organization_service.app.handle(
            command.UserCrudCommand(
                user=self._root["user"],
                objs=None,
                obj_ids=user_id,
                operation=CrudOperation.READ_ONE,
            )
        )
        return user

    def retrieve_user_permissions(self, user: model.User) -> set[Permission]:  # type: ignore[override]
        return self._rbac_service.retrieve_user_permissions(user)  # type: ignore[arg-type]



from __future__ import annotations

from gen_epix.seqdb.domain.service import BaseAbacService


class AbacService(BaseAbacService):

    def register_policies(self) -> None:
        """
        Placeholder for registering policies
        """
        pass



# pylint: disable=useless-import-alias
from gen_epix.casedb.domain.model.abac.abac import *
from gen_epix.casedb.domain.model.abac.case_abac import *
from gen_epix.casedb.domain.model.base import Model as Model
from gen_epix.casedb.domain.model.case.case import *
from gen_epix.casedb.domain.model.case.complete_case_type import *
from gen_epix.casedb.domain.model.geo import *
from gen_epix.casedb.domain.model.ontology import *
from gen_epix.casedb.domain.model.organization import *
from gen_epix.casedb.domain.model.seqdb import *
from gen_epix.casedb.domain.model.subject import *
from gen_epix.casedb.domain.model.system import *
from gen_epix.fastapp.services.auth import IdentityProvider as IdentityProvider
from gen_epix.fastapp.services.auth import IDPUser as IDPUser



from gen_epix.fastapp import BaseRepository
from gen_epix.omopdb.domain import DOMAIN
from gen_epix.omopdb.domain import model as model  # forces models to be registered now
from gen_epix.omopdb.domain.enum import ServiceType


class BaseSystemRepository(BaseRepository):
    ENTITIES = DOMAIN.get_dag_sorted_entities(
        service_type=ServiceType.SYSTEM, persistable=True
    )



import logging
from typing import Annotated, Type

from fastapi import Depends, Request, Security
from fastapi.security import SecurityScopes

from gen_epix.fastapp import App, exc, model
from gen_epix.fastapp.services.auth.base import BaseAuthService
from gen_epix.fastapp.services.auth.command import GetIdentityProvidersCommand
from gen_epix.fastapp.services.auth.idp_client import IDPClient
from gen_epix.fastapp.services.auth.mock_idp_client import MockIDPClient
from gen_epix.fastapp.services.auth.model import Claims, IdentityProvider, IDPUser
from gen_epix.fastapp.services.auth.util import create_idp_clients_from_config


class AuthService(BaseAuthService):
    def __init__(
        self,
        app: App,
        logger: logging.Logger | None = None,
        idps_cfg: list[dict[str, str | list]] | None = None,
        repository: None = None,
        **kwargs: dict,
    ):
        kwargs["service_type"] = kwargs.get(
            "service_type", BaseAuthService.SERVICE_TYPE
        )
        super().__init__(app, repository=repository, logger=logger, **kwargs)
        self._idp_clients: list[IDPClient] = []

        # Initialize authentication services
        self._idp_clients = create_idp_clients_from_config(app, idps_cfg)
        self._idp_client_by_id = {x.id: x for x in self._idp_clients or []}

        # Initialize no authentication user
        self._no_auth_user: model.User
        self._no_auth_idp_client: IDPClient = MockIDPClient(logger=logger)

    @property
    def idp_clients(self) -> list[IDPClient]:
        return list(self._idp_clients)

    def create_user_dependencies(
        self,
    ) -> tuple[Type[model.User], Type[model.User], Type[IDPUser]]:

        if not self._idp_clients:
            # No authentication -> create/retrieve root user
            user_manager = self.app.user_manager
            if not user_manager:
                raise exc.InitializationServiceError(
                    "No authentication services configured and no user generator provided"
                )
            self._no_auth_user = user_manager.create_root_user_from_claims({})

            async def dummy_get_existing_user(
                request: Request, _security_scopes: SecurityScopes
            ) -> model.User:
                claims = await self._no_auth_idp_client(request)
                if claims:
                    user = await self.get_existing_user_from_claims(
                        claims, request_userinfo=False
                    )
                    if user:
                        return user
                return self._no_auth_user

            async def dummy_get_new_user(
                request: Request, _security_scopes: SecurityScopes
            ) -> model.User:
                claims = await self._no_auth_idp_client(request)
                if claims:
                    user = await self.get_new_user_from_claims(
                        claims, request_userinfo=False
                    )
                    if user:
                        return user
                raise exc.UnauthorizedAuthError(
                    f"Unable to create user due to missing header or claims"
                )

            registered_user_dependency = Annotated[
                model.User,
                Security(dummy_get_existing_user, scopes=["openid", "profile"]),
            ]
            new_user_dependency = Annotated[
                model.User,
                Security(dummy_get_new_user, scopes=["openid", "profile"]),
            ]
            idp_user_dependency = Annotated[
                IDPUser,
                Security(dummy_get_new_user, scopes=["openid", "profile"]),
            ]

            return registered_user_dependency, new_user_dependency, idp_user_dependency

        # Init get_current_user function definition and environment
        # TODO: generate get_current_user and get_new_user functions
        # dynamically based on number of authentication services
        n_security_bases = len(self._idp_clients)
        if self._logger:
            self._logger.info(
                self.create_log_message(
                    "1e3d75cb",
                    "Generating user authentication function per security base",
                )
            )

        def _warn(request: Request) -> None:
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "f8853e9e",
                        "Unable to verify provided user",
                        request=request,
                    )
                )

        idp_client_list = self._idp_clients + ([None] * (5 - len(self._idp_clients)))

        async def get_current_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user1(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user2(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user3(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            if claims_3:
                return await self.get_existing_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            if claims_3:
                return await self.get_new_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user4(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            if claims_3:
                return await self.get_idp_user_from_claims(claims_3)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_current_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> model.User:
            if claims_0:
                return await self.get_existing_user_from_claims(claims_0)
            if claims_1:
                return await self.get_existing_user_from_claims(claims_1)
            if claims_2:
                return await self.get_existing_user_from_claims(claims_2)
            if claims_3:
                return await self.get_existing_user_from_claims(claims_3)
            if claims_4:
                return await self.get_existing_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_new_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> model.User:
            if claims_0:
                return await self.get_new_user_from_claims(claims_0)
            if claims_1:
                return await self.get_new_user_from_claims(claims_1)
            if claims_2:
                return await self.get_new_user_from_claims(claims_2)
            if claims_3:
                return await self.get_new_user_from_claims(claims_3)
            if claims_4:
                return await self.get_new_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        async def get_idp_user5(
            request: Request,
            _security_scopes: SecurityScopes,
            claims_0: Claims = Depends(idp_client_list[0]),
            claims_1: Claims = Depends(idp_client_list[1]),
            claims_2: Claims = Depends(idp_client_list[2]),
            claims_3: Claims = Depends(idp_client_list[3]),
            claims_4: Claims = Depends(idp_client_list[4]),
        ) -> IDPUser:
            if claims_0:
                return await self.get_idp_user_from_claims(claims_0)
            if claims_1:
                return await self.get_idp_user_from_claims(claims_1)
            if claims_2:
                return await self.get_idp_user_from_claims(claims_2)
            if claims_3:
                return await self.get_idp_user_from_claims(claims_3)
            if claims_4:
                return await self.get_idp_user_from_claims(claims_4)
            _warn(request)
            raise exc.UnauthorizedAuthError()

        get_idp_user_functions = [
            get_idp_user1,
            get_idp_user2,
            get_idp_user3,
            get_idp_user4,
            get_idp_user5,
        ]
        get_current_user_functions = [
            get_current_user1,
            get_current_user2,
            get_current_user3,
            get_current_user4,
            get_current_user5,
        ]
        get_new_user_functions = [
            get_new_user1,
            get_new_user2,
            get_new_user3,
            get_new_user4,
            get_new_user5,
        ]

        # Create CurrentUser/NewUser, injecting get_current_user/get_new_user
        if n_security_bases > len(get_current_user_functions):
            msg = (
                f"More than {len(get_current_user_functions)} "
                "({n_security_bases}) not implemented"
            )
            if self._logger:
                self._logger.error(self.create_log_message("d6f4ede7", msg))
            raise exc.InitializationServiceError(msg)
        registered_user_dependency: model.User = Annotated[  # type: ignore
            model.User,
            Security(
                get_current_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        new_user_dependency: model.User = Annotated[  # type: ignore
            model.User,
            Security(
                get_new_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        idp_user_dependency: IDPUser = Annotated[  # type: ignore
            IDPUser,
            Security(
                get_idp_user_functions[n_security_bases - 1],  # type: ignore
                scopes=["openid", "profile"],
            ),
        ]
        return registered_user_dependency, new_user_dependency, idp_user_dependency

    def get_identity_providers(
        self,
        _cmd: GetIdentityProvidersCommand,
    ) -> list[IdentityProvider]:
        return [x.get_identity_provider() for x in self._idp_clients]

    async def get_idp_user_from_claims(self, claims: Claims) -> IDPUser:
        claims_dict = claims.claims
        issuer: str = claims_dict["iss"]  # type: ignore
        sub: str = claims_dict["sub"]  # type: ignore

        return IDPUser(issuer=issuer, sub=sub)

    async def get_new_user_from_claims(
        self, claims: Claims, request_userinfo: bool = True
    ) -> model.User:
        # Get userinfo
        if request_userinfo:
            claims.claims.update(
                self._idp_client_by_id[claims.idp_client_id].get_claims_from_userinfo(
                    claims.token
                )
            )
        # Create user obj
        user_manager = self.app.user_manager
        if user_manager:
            # Use user manager to create user
            new_user = user_manager.get_user_instance_from_claims(claims.claims)
            if new_user:
                return new_user
        else:
            # No user manager configured, create user obj directly from claims
            new_user = model.User(**claims.claims)  # type: ignore
        return new_user

    async def get_existing_user_from_claims(
        self, claims: Claims, request_userinfo: bool = True
    ) -> model.User:
        issuer: str = claims.claims["iss"]  # type: ignore
        sub: str = claims.claims["sub"]  # type: ignore
        user_manager = self.app.user_manager
        if not user_manager:
            # No user generator configured
            raise exc.UnauthorizedAuthError()

        user_key = user_manager.get_user_key_from_claims(claims.claims)
        if not user_key and request_userinfo:
            claims.claims.update(
                self._idp_client_by_id[claims.idp_client_id].get_claims_from_userinfo(
                    claims.token
                )
            )
            user_key = user_manager.get_user_key_from_claims(claims.claims)
        if not user_key:
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "d3b7e9f1",
                        "No user key found in claims",
                        sub=sub,
                        user_key=user_key,
                    )
                )
            raise exc.UnauthorizedAuthError()

        try:
            user = user_manager.retrieve_user_by_key(user_key)
            return user
        except exc.NoResultsError:
            # User does not exist
            if self._logger:
                self._logger.warning(
                    self.create_log_message(
                        "ec9625ad",
                        "User not found",
                        issuer=issuer,
                        sub=sub,
                        user_key=user_key,
                    )
                )

            # Check if this is the root user, if so create it and and its dependencies
            if user_manager.is_root_user(claims.claims):
                if self._logger:
                    self._logger.warning(
                        self.create_log_message(
                            "d4f37b85",
                            "User is root user, creating",
                            issuer=issuer,
                            sub=sub,
                        )
                    )
                return user_manager.create_root_user_from_claims(claims.claims)

            # Automatically create the user if configured
            try:
                user = user_manager.create_user_from_claims(claims.claims)
                if not user:
                    raise exc.UnauthorizedAuthError()
                if self._logger and self._logger.level <= logging.DEBUG:
                    self._logger.debug(
                        self.create_log_message(
                            "fe8bfbd0",
                            "Automatically created user",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                        )
                    )
                return user
            except Exception as exception:
                if self._logger:
                    self._logger.error(
                        self.create_log_message(
                            "08e3c18b",
                            "Could not automatically create user",
                            issuer=issuer,
                            sub=sub,
                            user_key=user_key,
                            exception=exception,
                        )
                    )
                raise exc.UnauthorizedAuthError()



from typing import Callable

from fastapi import APIRouter, FastAPI

from gen_epix.casedb.api.base import EXCLUDED_PERMISSIONS
from gen_epix.casedb.domain import enum
from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator


def create_geo_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    registered_user_dependency: Callable | None = None,
    new_user_dependency: Callable | None = None,
    idp_user_dependency: Callable | None = None,
    handle_exception: Callable | None = None,
    **kwargs: dict,
) -> None:
    assert handle_exception
    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.GEO,
        user_dependency=registered_user_dependency,
        excluded_permissions=EXCLUDED_PERMISSIONS,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



import abc
import uuid
from typing import Type

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository.abac import BaseAbacRepository
from gen_epix.fastapp import BaseService
from gen_epix.fastapp.model import Command


class BaseAbacService(BaseService):
    SERVICE_TYPE = ServiceType.ABAC

    ORGANIZATION_ADMIN_WRITE_COMMANDS: set[Type[Command]] = {
        command.ContactCrudCommand,
        command.SiteCrudCommand,
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
    }

    UPDATE_USER_COMMANDS: set[Type[Command]] = {
        command.InviteUserCommand,
        command.UpdateUserCommand,
    }

    CASE_ABAC_COMMANDS: set[Type[Command]] = {
        # command.RetrieveCompleteUserCommand,
        command.RetrieveCompleteCaseTypeCommand,
        command.RetrieveCasesByQueryCommand,
        command.RetrieveCasesByIdCommand,
        command.RetrieveCaseRightsCommand,
        command.RetrieveCaseSetRightsCommand,
        command.RetrieveCaseTypeStatsCommand,
        command.RetrieveCaseSetStatsCommand,
        command.CaseTypeCrudCommand,
        command.CaseTypeSetMemberCrudCommand,
        command.CaseTypeSetCrudCommand,
        command.CaseTypeColCrudCommand,
        command.CaseTypeColSetCrudCommand,
        command.CaseTypeColSetMemberCrudCommand,
        command.CaseCrudCommand,
        # command.CaseDataCollectionUpdateAssociationCommand,
        command.CaseSetCreateCommand,
        command.CasesCreateCommand,
        command.CaseSetCrudCommand,
        # command.CaseSetCaseUpdateAssociationCommand,
        # command.CaseSetDataCollectionUpdateAssociationCommand,
        command.CaseDataCollectionLinkCrudCommand,
        command.CaseSetDataCollectionLinkCrudCommand,
        command.DataCollectionCrudCommand,
        command.RetrievePhylogeneticTreeByCasesCommand,
        command.RetrieveGeneticSequenceByCaseCommand,
        command.RetrieveCaseSetStatsCommand,
        command.RetrieveCaseTypeStatsCommand,
    }

    READ_ORGANIZATION_RESULTS_ONLY_COMMANDS: set[Type[Command]] = {
        command.UserCrudCommand,
        command.OrganizationAdminPolicyCrudCommand,
        command.OrganizationAccessCasePolicyCrudCommand,
        command.OrganizationShareCasePolicyCrudCommand,
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
        command.UserInvitationCrudCommand,
    }

    READ_SELF_RESULTS_ONLY_COMMANDS: set[Type[Command]] = {
        command.UserCrudCommand,
        command.UserAccessCasePolicyCrudCommand,
        command.UserShareCasePolicyCrudCommand,
    }

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseAbacRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)
        f(
            command.RetrieveOrganizationAdminNameEmailsCommand,
            self.retrieve_organization_admin_name_emails,
        )
        f(
            command.RetrieveCompleteUserCommand,
            self.retrieve_complete_user,
        )
        f(
            command.UpdateUserOwnOrganizationCommand,
            self.temp_update_user_own_organization,
        )

    @abc.abstractmethod
    def register_policies(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_complete_user(
        self, cmd: command.RetrieveCompleteUserCommand
    ) -> model.CompleteUser:
        raise NotImplementedError

    @abc.abstractmethod
    def retrieve_organization_admin_name_emails(
        self,
        cmd: command.RetrieveOrganizationAdminNameEmailsCommand,
    ) -> list[model.UserNameEmail]:
        raise NotImplementedError

    @abc.abstractmethod
    def temp_update_user_own_organization(
        self,
        cmd: command.UpdateUserOwnOrganizationCommand,
    ) -> model.User:
        raise NotImplementedError

    @abc.abstractmethod
    def get_case_abac(self, cmd: command.Command) -> model.CaseAbac:
        raise NotImplementedError

    @abc.abstractmethod
    def get_organizations_under_admin(self, user: model.User) -> set[uuid.UUID]:
        raise NotImplementedError
